import unittest

from p5.pmath.utils import (constrain, lerp, remap, normalize, magnitude, distance, sq)

class TestUtils(unittest.TestCase):

    def test_constrain(self):
        self.assertEqual(constrain(5, 0, 10), 5)
        self.assertEqual(constrain(-10, 0, 10), 0)
        self.assertEqual(constrain(20, 0, 10), 10)

    def test_lerp(self):
        self.assertEqual(lerp(0, 100, 0.5), 50)

    def test_remap(self):
        self.assertEqual(remap(50, (0, 100), (0, 10)), 5.0)

    def test_normalize(self):
        self.assertEqual(normalize(50, 0, 100), 0.5)

    def test_magnitude(self):
        self.assertEqual(magnitude(3, 4), 5)

    def test_distance(self):
        self.assertEqual(distance((0, 0, 0), (2, 3, 6)), 7)

    def test_sq(self):
        self.assertEqual(sq(4), 16)

if __name__ == "__main__":
    unittest.main()


#
# Part of p5: A Python package based on Processing
# Copyright (C) 2017-2019 Abhik Pal
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import colorsys
import math

from ..pmath import lerp
from ..pmath import constrain

from .constants import colour_codes

__all__ = ['color_mode', 'Color']

color_parse_mode = 'RGB'
color_range = (255, 255, 255, 255)

def color_mode(mode, max_1=255, max_2=None, max_3=None, max_alpha=255):
    """Set the color mode of the renderer.

    :param mode: One of {'RGB', 'HSB'} corresponding to Red/Green/Blue
        or Hue/Saturation/Brightness
    :type mode: str

    :param max_1: Maximum value for the first color channel (default:
        255)
    :type max_1: int

    :param max_2: Maximum value for the second color channel (default:
        max_1)
    :type max_2: int

    :param max_3: Maximum value for the third color channel (default:
        max_1)
    :type max_3: int

    :param max_alpha: Maximum valu for the alpha channel (default:
        255)
    :type max_alpha: int

    """
    global color_parse_mode
    global color_range

    if max_2 is None:
        max_2 = max_1

    if max_3 is None:
        max_3 = max_1

    color_range = (max_1, max_2, max_3, max_alpha)
    color_parse_mode = mode

def parse_color(*args, color_mode='RGB', normed=False, **kwargs):
    """Parses a color from a range of different input formats.

    This assumes that the args and kwargs are in the following form:

    - gray
    - gray, alpha = ...
    - gray, alpha
    - r, g, b
    - h, s, v
    - r, g, b, a
    - h, s, v, a
    - hex
    - colour name

    - gray = ...
    - gray = ..., alpha = ...
    - r = ..., g = ..., b = ...,
    - red = ..., green = ..., blue = ...,
    - h = ..., s = ..., b = ...,
    - hue = ..., saturation = ..., brightness = ...,
    - r = ..., g = ..., b = ..., a = ...
    - red = ..., green = ..., blue = ..., alpha = ...
    - h = ..., s = ..., b = ..., a = ...
    - hue = ..., saturation = ..., brightness = ..., alpha = ...

    :param args: The positional arguments that define the color.
    :type args: tuple

    :param kwargs: The keyword arguments that define the color.
    :type kwargs: dict

    :returns: The color parsed as red, green, blue, alpha values.
    :rtype: tuple

    """

    if 'alpha' in kwargs:
        alpha = kwargs['alpha']
    elif 'a' in kwargs:
        alpha = kwargs['a']
    else:
        alpha = 1 if normed else color_range[3]

    hsb = None
    rgb = None

    if len(args) == 1:
        if isinstance(args[0], int) or isinstance(args[0], float):
            gray = args[0]
            rgb = gray, gray, gray
        elif isinstance(args[0], str):
            name = args[0].lower()
            if name == "none":
                alpha = 0
                rgb = (0, 0, 0)
            else:
                if name[0] == "#":
                    hexadecimal = args[0]                
                elif name in colour_codes.keys():
                    hexadecimal = colour_codes[name]
                else:
                    raise ValueError("Invalid colour name %s" %name)

                alpha = 255
                _r = int(hexadecimal[1:3], 16)
                _g = int(hexadecimal[3:5], 16)
                _b = int(hexadecimal[5:7], 16)
                rgb = (_r, _g, _b)

    elif len(args) == 2:
        gray, alpha = args
        rgb =  gray, gray, gray
    elif (len(args) == 3) and color_mode.startswith('RGB'):
        rgb = args
    elif (len(args) == 3) and color_mode.startswith('HSB'):
        hsb = args
    elif (len(args) == 4) and color_mode.startswith('RGB'):
        _r, _g, _b, alpha = args
        rgb = (_r, _g, _b)
    elif (len(args) == 4) and color_mode.startswith('HSB'):
        _h, _s, _b, alpha = args
        hsb = (_h, _s, _b)
    elif 'gray' in kwargs:
        gray = kwargs['gray']
        rgb = gray, gray, gray
    elif all(param in kwargs for param in ['red', 'green', 'blue']):
        _r = kwargs['red']
        _g = kwargs['green']
        _b = kwargs['blue']
        rgb = (_r, _g, _b)
    elif all(param in kwargs for param in ['r', 'g', 'b']):
        _r = kwargs['r']
        _g = kwargs['g']
        _b = kwargs['b']
        rgb = (_r, _g, _b)
    elif all(param in kwargs for param in ['hue', 'saturation', 'brightness']):
        _h = kwargs['hue']
        _s = kwargs['saturation']
        _b = kwargs['brightness']
        hsb = (_h, _s, _b)
    elif all(param in kwargs for param in ['h', 's', 'b']):
        _h = kwargs['h']
        _s = kwargs['s']
        _b = kwargs['b']
        hsb = (_h, _s, _b)
    else:
        raise ValueError("Failed to parse color.")

    if not (hsb is None):
        h, s, b = hsb
        if not normed:
            h = constrain(h / color_range[0], 0, 1)
            s = constrain(s / color_range[1], 0, 1)
            b = constrain(b / color_range[2], 0, 1)
        red, green, blue = colorsys.hsv_to_rgb(h, s, b)

    if not (rgb is None):
        r, g, b = rgb
        if not normed:
            red = constrain(r / color_range[0], 0, 1)
            green = constrain(g / color_range[1], 0, 1)
            blue = constrain(b / color_range[2], 0, 1)
        else:
            red, green, blue = r, g, b

    if not normed:
        alpha = constrain(alpha / color_range[3], 0, 1)

    return red, green, blue, alpha


class Color:
    """Represents a color."""
    def __init__(self, *args, color_mode=None, normed=False, **kwargs):
        if color_mode is None:
            color_mode = color_parse_mode

        if (len(args) == 1) and isinstance(args[0], Color):
            r = args[0]._red
            g = args[0]._green
            b = args[0]._blue
            a = args[0]._alpha
        elif len(args) == 2 and isinstance(args[0], Color):
            r = args[0]._red
            g = args[0]._green
            b = args[0]._blue
            a = args[1]
        else:
            r, g, b, a = parse_color(*args, color_mode=color_mode,
                                     normed=normed, **kwargs)

        self._red = r
        self._green = g
        self._blue = b
        self._alpha = a

        self._recompute_hsb()

    def _recompute_rgb(self):
        """Recompute the RGB values from HSB values."""
        r, g, b = colorsys.hsv_to_rgb(self._hue, self._saturation, self._brightness)
        self._red = r
        self._greeen = g
        self._blue = b

    def _recompute_hsb(self):
        """Recompute the HSB values from the RGB values."""
        h, s, b = colorsys.rgb_to_hsv(self._red, self._green, self._blue)
        self._hue = h
        self._saturation = s
        self._brightness = b

    def lerp(self, target, amount):
        """Linearly interpolate one color to another by the given amount.

        :param target: The target color to lerp to.
        :type target: Color

        :param amount: The amount by which the color should be lerped
            (should be a float between 0 and 1).
        :type amount: float

        :returns: A new color lerped between the current color and the
            other color.
        :rtype: Color

        """
        lerped = (lerp(s, t, amount) for s, t in zip(self.rgba, target.rgba))
        return Color(*lerped, color_mode='RGB')

    def __repr__(self):
        fvalues = self._red, self._green, self._blue
        return "Color( red={}, green={}, blue={} )".format(*fvalues)

    __str__ = __repr__

    def __eq__(self, other):
        return all(math.isclose(sc, oc)
                   for sc, oc in zip(self.normalized, other.normalized))

    def __neq__(self, other):
        return not all(math.isclose(sc, oc)
                       for sc, oc in zip(self.normalized, other.normalized))

    @property
    def normalized(self):
        """Normalized RGBA color values"""
        return (self._red, self._green, self._blue, self._alpha)

    @property
    def gray(self):
        """The gray-scale value of the color.

        Performs a luminance conversion of the current color to
        grayscale.

        """
        # The formula we use to convert to grayscale is approximate
        # and probably not as accurate as a proper coloremetric
        # conversion. However, the number of calculations required is
        # less and GIMP uses something similar so we should be fine.
        #
        # REFERENCES:
        #
        # - "Converting Color Images to B&W"
        #   <https://www.gimp.org/tutorials/Color2BW/>
        #
        # - "Luma Coding in Video Systems" from "Grayscale"
        #   <https://en.wikipedia.org/wiki/Grayscale#Luma_coding_in_video_systems>
        #
        # - Wikipedia : Grayscale
        #   <https://en.wikipedia.org/wiki/Grayscale#Converting_color_to_grayscale>
        #
        # - Conversion to grayscale, sample implementation (StackOverflow)
        # <https://stackoverflow.com/a/15686412>
        norm_gray  = 0.299 * self._red + 0.587 * self._green + 0.144 * self._blue
        return norm_gray * 255

    @gray.setter
    def gray(self, value):
        value = constrain(value / 255, 0, 1)
        self._red = value
        self._green = value
        self._blue = value
        self._recompute_hsb()

    @property
    def alpha(self):
        """The alpha value for the color."""
        return self._alpha * color_range[3]

    @alpha.setter
    def alpha(self, value):
        self._alpha = constrain(value / color_range[3], 0, 1)

    @property
    def rgb(self):
        """
        :returns: Color components in RGB.
        :rtype: tuple
        """
        return (self.red, self.green, self.blue)

    @property
    def rgba(self):
        """
        :returns: Color components in RGBA.
        :rtype: tuple
        """
        return (self.red, self.green, self.blue, self.alpha)


    @property
    def red(self):
        """The red component of the color"""
        return self._red * color_range[0]

    @red.setter
    def red(self, value):
        self._red = constrain(value / color_range[0], 0, 1)
        self._recompute_hsb()

    @property
    def green(self):
        """The green component of the color"""
        return self._green * color_range[1]

    @green.setter
    def green(self, value):
        self._green = constrain(value / color_range[1], 0, 1)
        self._recompute_hsb()

    @property
    def blue(self):
        """The blue component of the color"""
        return self._blue * color_range[2]

    @blue.setter
    def blue(self, value):
        self._blue = constrain(value / color_range[2], 0, 1)
        self._recompute_hsb()

    @property
    def hsb(self):
        """
        :returns: Color components in HSB.
        :rtype: tuple
        """
        return (self._hue, self._saturation, self._brightness)

    @property
    def hsba(self):
        """
        :returns: Color components in HSBA.
        :rtype: tuple
        """
        return (self.hue, self.saturation, self.brightness, self.alpha)

    @property
    def hue(self):
        """The hue component of the color"""
        return self._hue * color_range[0]

    @hue.setter
    def hue(self, value):
        self._hue = constrain(value / color_range[0], 0, 1)
        self._recompute_rgb()

    @property
    def saturation(self):
        """The saturation component of the color"""
        return self._saturation * color_range[1]

    @saturation.setter
    def saturation(self, value):
        self._saturation = constrain(value / color_range[1], 0, 1)
        self._recompute_rgb()

    @property
    def brightness(self):
        """The brightness component of the color"""
        return self._brightness * color_range[2]

    @brightness.setter
    def brightness(self, value):
        self._brightness = constrain(value / color_range[2], 0, 1)
        self._recompute_rgb()

    # ...and some convenient aliases
    r = red
    g = green
    h = hue
    s = saturation
    value = brightness
    v = value

    # `b` is tricky. depending on the current color code, this could
    # either be the brightness value or the blue value.
    @property
    def b(self):
        """The blue or the brightness value (depending on the color mode)."""
        if color_parse_mode== 'RGB':
            return self.blue
        elif color_parse_mode== 'HSB':
            return self.brightness
        else:
            raise ValueError("Unknown color mode {}".format(color_parse_mode))

    @b.setter
    def b(self, value):
        if color_parse_mode== 'RGB':
            self.blue = value
        elif color_parse_mode== 'HSB':
            self.brightness = value
        else:
            raise ValueError("Unknown color mode {}".format(color_parse_mode))

    @property
    def hex(self):
        """
        :returns: Color as a hex value
        :rtype: str
        """
        raise NotImplementedError()



#
# Part of p5: A Python package based on Processing
# Copyright (C) 2017-2019 Abhik Pal
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

"""Constants Variables for P5 sketches"""

# Environment modes
P2D = "P2D"
P3D = "P3D"

# Stroke Cap Properties
ROUND = "ROUND"
SQUARE = "SQUARE"
PROJECT = "PROJECT"

# Stroke Join Properties
MITER = "MITER"
BEVEL = "BEVEL"
ROUND = "ROUND"

# Ellipse and Rectangle Modes
RADIUS = "RADIUS"
CENTER = "CENTER"
CORNER = "CORNER"
CORNERS = "CORNERS"

# Colors
colour_codes = {
	'indianred': '#CD5C5C',
	'lightcoral': '#F08080',
	'salmon': '#FA8072',
	'darksalmon': '#E9967A',
	'lightsalmon': '#FFA07A',
	'crimson': '#DC143C',
	'red': '#FF0000',
	'firebrick': '#B22222',
	'darkred': '#8B0000',
	'pink': '#FFC0CB',
	'lightpink': '#FFB6C1',
	'hotpink': '#FF69B4',
	'deeppink': '#FF1493',
	'mediumvioletred': '#C71585',
	'palevioletred': '#DB7093',
	'lightsalmon': '#FFA07A',
	'coral': '#FF7F50',
	'tomato': '#FF6347',
	'orangered': '#FF4500',
	'darkorange': '#FF8C00',
	'orange': '#FFA500',
	'gold': '#FFD700',
	'yellow': '#FFFF00',
	'lightyellow': '#FFFFE0',
	'lemonchiffon': '#FFFACD',
	'lightgoldenrodyellow': '#FAFAD2',
	'papayawhip': '#FFEFD5',
	'moccasin': '#FFE4B5',
	'peachpuff': '#FFDAB9',
	'palegoldenrod': '#EEE8AA',
	'khaki': '#F0E68C',
	'darkkhaki': '#BDB76B',
	'lavender': '#E6E6FA',
	'thistle': '#D8BFD8',
	'plum': '#DDA0DD',
	'violet': '#EE82EE',
	'orchid': '#DA70D6',
	'fuchsia': '#FF00FF',
	'magenta': '#FF00FF',
	'mediumorchid': '#BA55D3',
	'mediumpurple': '#9370DB',
	'rebeccapurple': '#663399',
	'blueviolet': '#8A2BE2',
	'darkviolet': '#9400D3',
	'darkorchid': '#9932CC',
	'darkmagenta': '#8B008B',
	'purple': '#800080',
	'indigo': '#4B0082',
	'slateblue': '#6A5ACD',
	'darkslateblue': '#483D8B',
	'mediumslateblue': '#7B68EE',
	'greenyellow': '#ADFF2F',
	'chartreuse': '#7FFF00',
	'lawngreen': '#7CFC00',
	'lime': '#00FF00',
	'limegreen': '#32CD32',
	'palegreen': '#98FB98',
	'lightgreen': '#90EE90',
	'mediumspringgreen': '#00FA9A',
	'springgreen': '#00FF7F',
	'mediumseagreen': '#3CB371',
	'seagreen': '#2E8B57',
	'forestgreen': '#228B22',
	'green': '#008000',
	'darkgreen': '#006400',
	'yellowgreen': '#9ACD32',
	'olivedrab': '#6B8E23',
	'olive': '#808000',
	'darkolivegreen': '#556B2F',
	'mediumaquamarine': '#66CDAA',
	'darkseagreen': '#8FBC8B',
	'lightseagreen': '#20B2AA',
	'darkcyan': '#008B8B',
	'teal': '#008080',
	'aqua': '#00FFFF',
	'cyan': '#00FFFF',
	'lightcyan': '#E0FFFF',
	'paleturquoise': '#AFEEEE',
	'aquamarine': '#7FFFD4',
	'turquoise': '#40E0D0',
	'mediumturquoise': '#48D1CC',
	'darkturquoise': '#00CED1',
	'cadetblue': '#5F9EA0',
	'steelblue': '#4682B4',
	'lightsteelblue': '#B0C4DE',
	'powderblue': '#B0E0E6',
	'lightblue': '#ADD8E6',
	'skyblue': '#87CEEB',
	'lightskyblue': '#87CEFA',
	'deepskyblue': '#00BFFF',
	'dodgerblue': '#1E90FF',
	'cornflowerblue': '#6495ED',
	'mediumslateblue': '#7B68EE',
	'royalblue': '#4169E1',
	'blue': '#0000FF',
	'mediumblue': '#0000CD',
	'darkblue': '#00008B',
	'navy': '#000080',
	'midnightblue': '#191970',
	'cornsilk': '#FFF8DC',
	'blanchedalmond': '#FFEBCD',
	'bisque': '#FFE4C4',
	'navajowhite': '#FFDEAD',
	'wheat': '#F5DEB3',
	'burlywood': '#DEB887',
	'tan': '#D2B48C',
	'rosybrown': '#BC8F8F',
	'sandybrown': '#F4A460',
	'goldenrod': '#DAA520',
	'darkgoldenrod': '#B8860B',
	'peru': '#CD853F',
	'chocolate': '#D2691E',
	'saddlebrown': '#8B4513',
	'sienna': '#A0522D',
	'brown': '#A52A2A',
	'maroon': '#800000',
	'white': '#FFFFFF',
	'snow': '#FFFAFA',
	'honeydew': '#F0FFF0',
	'mintcream': '#F5FFFA',
	'azure': '#F0FFFF',
	'aliceblue': '#F0F8FF',
	'ghostwhite': '#F8F8FF',
	'whitesmoke': '#F5F5F5',
	'seashell': '#FFF5EE',
	'beige': '#F5F5DC',
	'oldlace': '#FDF5E6',
	'floralwhite': '#FFFAF0',
	'ivory': '#FFFFF0',
	'antiquewhite': '#FAEBD7',
	'linen': '#FAF0E6',
	'lavenderblush': '#FFF0F5',
	'mistyrose': '#FFE4E1',
	'gainsboro': '#DCDCDC',
	'lightgray': '#D3D3D3',
	'silver': '#C0C0C0',
	'darkgray': '#A9A9A9',
	'gray': '#808080',
	'dimgray': '#696969',
	'lightslategray': '#778899',
	'slategray': '#708090',
	'darkslategray': '#2F4F4F',
	'black': '#000000',
}



import unittest
import numpy as np

from p5.pmath import matrix
from p5.pmath.utils import PI

v = np.array([3, 4])

class TestMatrix(unittest.TestCase):
    
    def test_magnitude(self):
        self.assertEqual(matrix._magnitude(v), 5.0)

    def test_normalize(self):
        self.assertTrue(np.array_equal(matrix._normalize(v), v/5.0))

    def test_scale_transform(self):
        T = matrix.scale_transform(1, 2, 3)
        R = np.array([
            [1, 0, 0, 0],
            [0, 2, 0, 0],
            [0, 0, 3, 0],
            [0, 0, 0, 1]
            ])
        self.assertTrue(np.array_equal(T, R))

    def test_translation_matrix(self):
        T = matrix.translation_matrix(5, 6, 7)
        R = np.array([
            [1, 0, 0, 5],
            [0, 1, 0, 6],
            [0, 0, 1, 7],
            [0, 0, 0, 1]
            ])
        self.assertTrue(np.array_equal(T, R))

    def test_rotation_matrix(self):
        axis = np.array([1, 1, 1])
        T = matrix.rotation_matrix(axis, PI/2)
        R = np.array([
            [ 0.33333333, -0.24401694, 0.9106836, 0],
            [ 0.9106836, 0.33333333, -0.24401694, 0],
            [-0.24401694, 0.9106836, 0.33333333, 0],
            [ 0, 0, 0, 1]
            ])
        self.assertTrue(np.allclose(T, R))

    def test_triple_axis_rotation_matrix(self):
        x = np.array([1, 1, 0])
        y = np.array([0, 1, 0])
        z = np.array([0, 1, 1])

        T = matrix.triple_axis_rotation_matrix(x, y, z)
        R = np.array([
            [1, 1, 0, 0],
            [0, 1, 0, 0],
            [0, 1, 1, 0],
            [0, 0, 0, 1],
            ])
        self.assertTrue(np.array_equal(T, R))

    def test_look_at(self):
        eye = np.array([0, 0, 10])
        at = np.array([0, 0, 0])
        up = np.array([0, 1, 0])

        T = matrix.look_at(eye, at, up)
        R = np.array([
            [1, 0, 0, 0],
            [0, 1, 0, 0],
            [0, 0, 1, -10],
            [0, 0, 0, 1]
            ])
        self.assertTrue(np.array_equal(T, R))

    def test_perspective_matrix(self):
        fov = 2/3*PI
        aspect = 3/4
        near_plane = 0.1
        far_plane = 100

        T = matrix.perspective_matrix(fov, aspect, near_plane, far_plane)
        R = np.array([
            [0.769800, 0, 0, 0],
            [0, 0.5773502, 0, 0],
            [0, 0, -1.00200, -0.200200],
            [0, 0, -1, 0]
            ])

        self.assertTrue(np.allclose(T, R))

if __name__ == "__main__":
    unittest.main()


#
# Part of p5: A Python package based on Processing
# Copyright (C) 2017-2019 Abhik Pal
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import builtins

from .color import Color
from .image import image
from .image import image_mode
from .image import PImage
from .structure import push_style
from .transforms import push_matrix
from .constants import *

from . import p5

__all__ = [ 'background', 'clear', 'fill', 'no_fill',
            'stroke', 'no_stroke', 'tint', 'no_tint' , 
            'stroke_weight', 'stroke_cap', 'stroke_join']

stroke_cap_codes = {
    'PROJECT': 0,
    'SQUARE': 1,
    'ROUND': 2
}

stroke_join_codes = {
    'MITER': 0,
    'BEVEL': 1,
    'ROUND': 2
}

def fill(*fill_args, **fill_kwargs):
    """Set the fill color of the shapes.

    :param fill_args: positional arguments to be parsed as a color.
    :type fill_args: tuple

    :param fill_kwargs: keyword arguments to be parsed as a color.
    :type fill_kwargs: dict

    :returns: The fill color.
    :rtype: Color

    """
    fill_color = Color(*fill_args, **fill_kwargs)
    p5.renderer.fill_enabled = True
    p5.renderer.fill_image_enabled = False
    p5.renderer.fill_color = fill_color.normalized
    return fill_color

def no_fill():
    """Disable filling geometry."""
    p5.renderer.fill_enabled = False

def stroke(*color_args, **color_kwargs):
    """Set the color used to draw lines around shapes

    :param color_args: positional arguments to be parsed as a color.
    :type color_args: tuple

    :param color_kwargs: keyword arguments to be parsed as a color.
    :type color_kwargs: dict

    :note: Both color_args and color_kwargs are directly sent to
        Color.parse_color

    :returns: The stroke color.
    :rtype: Color
    """
    stroke_color = Color(*color_args, **color_kwargs)
    p5.renderer.stroke_enabled = True
    p5.renderer.stroke_color = stroke_color.normalized

def stroke_weight(thickness):
    """Sets the width of the stroke used for lines, points, and the border around shapes. All widths are set in units of pixels.

    :param weight: thickness of stroke in pixels
    :type weight: int

    """
    p5.renderer.stroke_weight = thickness

def no_stroke():
    """Disable drawing the stroke around shapes."""
    p5.renderer.stroke_enabled = False

def stroke_cap(c):
    """Sets the style of line endings. The ends are SQUARE, 
    PROJECT, and ROUND. The default cap is ROUND.

    :param c: either 'SQUARE', 'PROJECT' or 'ROUND'
    :type c: string

    """
    if c in stroke_cap_codes.keys():
        p5.renderer.stroke_cap = stroke_cap_codes[c]
    else:
        raise ValueError("Invalid Stroke Cap %s" % c)

def stroke_join(j):
    """Sets the style of the joints which connect line segments. 
    These joints are either mitered, beveled, or rounded and 
    specified with the corresponding parameters MITER, BEVEL, 
    and ROUND. The default joint is MITER.

    :param weight: either 'MITER', 'BEVEL' or 'ROUND'
    :type j: string

    """
    if j in stroke_join_codes.keys():
        p5.renderer.stroke_join = stroke_join_codes[j]
    else:
        raise ValueError("Invalid Stroke Cap %s" % j)

def tint(*color_args, **color_kwargs):
    """Set the tint color for the sketch.

    :param color_args: positional arguments to be parsed as a color.
    :type color_args: tuple

    :param color_kwargs: keyword arguments to be parsed as a color.
    :type color_kwargs: dict

    :note: Both color_args and color_kwargs are directly sent to
        Color.parse_color

    :returns: The tint color.
    :rtype: Color
    """
    tint_color = Color(*color_args, **color_kwargs)
    p5.renderer.tint_enabled = True
    p5.renderer.tint_color = tint_color.normalized

def no_tint():
    """Disable tinting of images."""
    p5.renderer.tint_enabled = False

def background(*args, **kwargs):
    """Set the background color for the p5.renderer.

    :param args: positional arguments to be parsed as a color.
    :type color_args: tuple

    :param kwargs: keyword arguments to be parsed as a color.
    :type kwargs: dict

    :note: Both args and color_kwargs are directly sent to
        color.parse_color

    :note: When setting an image as the background, the dimensions of
        the image should be the same as that of the sketch window.

    :returns: The background color or image.
    :rtype: p5.Color | p5.PImage

    :raises ValueError: When the dimensions of the image and the
        sketch do not match.

    """
    if len(args) == 1 and isinstance(args[0], PImage):
        background_image = args[0]
        sketch_size = (builtins.width, builtins.height)

        if sketch_size != background_image.size:
            msg = "Image dimension {} and sketch dimension {} do not match"
            raise ValueError(msg.format(background_image.size, sketch_size))

        with push_style():
            no_tint()
            image_mode('corner')
            with push_matrix():
                image(background_image, (0, 0))

        return background_image

    with push_style():
        background_color = Color(*args, **kwargs)
        fill(background_color)
        no_stroke()

        with push_matrix():
            p5.renderer.background_color = background_color.normalized
            p5.renderer.clear()

def clear():
    """
    Clears the pixels within a buffer. 
    """
    p5.renderer.clear()



#
# Part of p5: A Python package based on Processing
# Copyright (C) 2017-2019 Abhik Pal
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

from collections import namedtuple
import numpy as np

__all__ = ['Vector', 'Point']

# Floating point precision for vectors.
EPSILON = 1e-8

Point = namedtuple('Point', ['x', 'y', 'z'])
Point.__new__.__defaults__ = (None, None, 0)

class Vector(Point):
    """Describes a vector in two or three dimensional space.

    A Vector -- specifically an Euclidean (or geometric) vector -- in
    two or three dimensional space is a geometric entity that has some
    magnitude (or length) and a direction.

    Examples::

        >>> vec_2d = Vector(3, 4)
        >>> vec_2d
        Vector(3.00, 4.00, 0.00)

        >>> vec_3d = Vector(2, 3, 4)
        >>> vec_3d
        Vector(2.00, 3.00, 4.00)

    :param x: The x-component of the vector.
    :type x: int or float

    :param y: The y-component of the vector.
    :type y: int or float

    :param z: The z-component of the vector (0 by default; only
        required for 3D vectors; )
    :type z: int or float

    """
    def __init__(self, x, y, z=0):
        self._array = np.array([x, y, z], dtype=np.float32)

    @property
    def x(self):
        """The x-component of the point."""
        return self._array[0]

    @x.setter
    def x(self, value):
        self._array[0] = value

    @property
    def y(self):
        """The y-component of the point."""
        return self._array[1]

    @y.setter
    def y(self, value):
        self._array[1] = value

    @property
    def z(self):
        """The z-component of the point."""
        return self._array[2]

    @z.setter
    def z(self, value):
        self._array[2] = value

    def distance(self, other):
        """Return the distance between two points.

        :returns: The distance between the current point and the given
            point.
        :rtype: float

        """
        return np.sqrt(np.sum((self._array - other._array) ** 2))

    dist = distance

    def lerp(self, other, amount):
        """Linearly interpolate from one point to another.

        :param other: Point to be interpolate to.

        :param amount: Amount by which to interpolate.
        :type amount: float

        :returns: Vector obtained by linearly interpolating this
            vector to the other vector by the given amount.

        """
        x, y, z = self._array + amount * (other._array - self._array)
        return self.__class__(x, y, z)

    def __add__(self, other):
        """Add the location of one point to that of another.

        Examples::

            >>> p = Vector(2, 3, 6)
            >>> q = Vector(3, 4, 5)
            >>> p + q
            Vector(5.00, 7.00, 11.00)

        :param other:

        :returns: The point obtained by adding the corresponding
            components of the two vectors.

        """
        x, y, z = self._array + other._array
        return self.__class__(x, y, z)

    def __sub__(self, other):
        """Subtract the location of one point from that of another.

        Examples::

            >>> p = Vector(2, 3, 6)
            >>> q = Vector(3, 4, 5)
            >>> p - q
            Vector(-1.00, -1.00, 1.00)

        :param other:
        :returns: The vector obtained by subtracteing  the corresponding
            components of the vector from those of another.

        """
        x, y, z = self._array - other._array
        return self.__class__(x, y, z)

    def __mul__(self, k):
        """Multiply the point by a scalar.

        Examples::

            >>> p = Vector(2, 3, 6)
            >>> p * 2
            Vector(4.00, 6.00, 12.00)

            >>> 2 * p
            Vector(4.00, 6.00, 12.00)

            >>> p * p
            Traceback (most recent call last):
                ...
            TypeError: Can't multiply/divide a point by a non-numeric.

            >>> p = Vector(2, 3, 6)
            >>> -p
            Vector(-2.00, -3.00, -6.00)

            >>> p = Vector(2, 3, 6)
            >>> p / 2
            Vector(1.00, 1.50, 3.00)

        :param k:
        :type k: int, float
        :returns: The vector obtained by multiplying each component of
            `self` by k.

        :raises TypeError: When `k` is non-numeric.

        """
        if isinstance(k, int) or isinstance(k, float):
            x, y, z = k * self._array
            return self.__class__(x, y, z)
        raise TypeError("Can't multiply/divide a point by a non-numeric.")

    def __rmul__(self, other):
        return self * other

    def __neg__(self):
        """Negate the vector."""
        return (-1) * self

    def __truediv__(self, other):
        """Divide the vector by a scalar."""
        return self * (1 / other)

    def cross(self, other):
        """Return the cross product of the two vectors.

        Examples::

            >>> i = Vector(1, 0, 0)
            >>> j = Vector(0, 1, 0)
            >>> i.cross(j)
            Vector(0.00, 0.00, 1.00)

        :param other:
        :type other: Vector
        :returns: The vector perpendicular to both `self` and `other`
            i.e., the vector obtained by taking the cross product of
            `self` and `other`.
        :rtype: Vector

        """
        x, y, z = np.cross(self._array, other._array)
        return self.__class__(x, y, z)

    def dot(self, other):
        """Compute the dot product of two vectors.

        Examples::

            >>> p = Vector(2, 3, 6)
            >>> q = Vector(3, 4, 5)
            >>> p.dot(q)
            48
            >>> p @ q
            48

        :param other:
        :type other: Vector
        :returns: The dot product of the two vectors.
        :rtype: int or float

        """
        return np.dot(self._array, other._array)

    @property
    def angle(self):
        """The angle of rotation of the vector (in radians).

        This attribute isn't available for three dimensional vectors.


        Examples::

            >>> from math import pi, isclose
            >>> p = Vector(1, 0, 0)
            >>> isclose(p.angle, 0)
            True

            >>> p = Vector(0, 1, 0)
            >>> isclose(p.angle, pi/2)
            True

            >>> p = Vector(1, 1, 1)
            >>> p.angle
            Traceback (most recent call last):
                ...
            ValueError: Can't compute the angle for a 3D vector.

            >>> p = Vector(1, 1)
            >>> isclose(p.angle, pi/4)
            True
            >>> p.angle = pi/2
            >>> isclose(p.angle, pi/2)
            True

            >>> p = Vector(1, 1)
            >>> isclose(p.angle, pi/4)
            True
            >>> p.rotate(pi/4)
            >>> isclose(p.angle, pi/2)
            True

        :raises ValueError: If the vector is three-dimensional

        """
        if np.abs(self.z) > EPSILON:
            raise ValueError("Can't compute the angle for a 3D vector.")
        return np.arctan2(self.y, self.x)

    @angle.setter
    def angle(self, theta):
        self.rotate(theta -  self.angle)

    def rotate(self, theta):
        """Rotates the vector by an angle.

        :param theta: Angle (in radians).
        :type theta: float or int

        """
        x = self.x * np.cos(theta) - self.y * np.sin(theta)
        y = self.x * np.sin(theta) + self.y * np.cos(theta)
        self.x = x
        self.y = y

    def angle_between(self, other):
        """Calculate the angle between two vectors.

        Examples::

            >>> from math import degrees
            >>> k = Vector(0, 1)
            >>> j = Vector(1, 0)
            >>> degrees(k.angle_between(j))
            90.0

        :param other:
        :type other: Vector

        :returns: The angle between `self` and `other` (in radians)
        :rtype: float

        """
        return np.arccos((np.dot(self._array, other._array)) /
                         (self.magnitude * other.magnitude))

    @property
    def magnitude(self):
        """The magnitude of the vector.

        Examples::

            >>> p = Vector(2, 3, 6)
            >>> p.magnitude
            7.0

            >>> abs(p)
            7.0

            >>> p.magnitude = 14
            >>> p
            Vector(4.00, 6.00, 12.00)

            >>> p.normalize()
            >>> print(p)
            Vector(0.29, 0.43, 0.86)

        """
        return np.sqrt(np.dot(self._array, self._array))

    @magnitude.setter
    def magnitude(self, new_magnitude):
        current_magnitude = self.magnitude
        self._array = (new_magnitude * self._array) / current_magnitude

    @property
    def magnitude_sq(self):
        """The squared magnitude of the vector."""
        return np.dot(self._array, self._array)

    @magnitude_sq.setter
    def magnitude_sq(self, new_magnitude_sq):
        self.magnitude = np.sqrt(new_magnitude_sq)

    def __abs__(self):
        """Return the magnitude of the vector."""
        return self.magnitude

    def normalize(self):
        """Set the magnitude of the vector to one."""
        if self.magnitude == 0.0:
            raise ValueError("Vector has magnitude 0; can't normalize.")
        self.magnitude = 1
        return self

    def limit(self, upper_limit=None, lower_limit=None):
        """Limit the magnitude of the vector to the given range.

        :param upper_limit: The upper limit for the limiting range
            (defaults to None).
        :type upper_limit: float

        :param lower_limit: The lower limit for the limiting range
            (defaults to None).
        :type lower_limit: float

        """
        magnitude = self.magnitude
        if upper_limit is None:
            upper_limit = magnitude
        if lower_limit is None:
            lower_limit = magnitude

        if magnitude < lower_limit:
            self.magnitude = lower_limit
        elif magnitude > upper_limit:
            self.magnitude = upper_limit

    def __matmul__(self, other):
        return self.dot(other)

    @classmethod
    def from_angle(cls, angle):
        """Return a new unit vector with the given angle.

        :param angle: Angle to be used to create the vector (in
            radians).
        :type angle: float
        """
        vec = cls.random_2D()
        vec.angle = angle
        return vec

    @classmethod
    def random_2D(cls):
        """Return a random 2D unit vector.
        """
        x, y = 2 * (np.random.random(2) - 0.5)
        vec = cls(x, y)
        vec.normalize()
        return vec

    @classmethod
    def random_3D(cls):
        """Return a new random 3D unit vector.
        """
        x, y, z = np.random.random(3)
        vec = cls(x, y, z)
        vec.normalize()
        return vec


    def copy(self):
        """Return a copy of the current point.

        :returns: A copy of the current point.
        :rtype: Vector

        """
        x, y, z = self._array
        return self.__class__(x, y, z)

    def __setitem__(self, key, value):
        self._array[key] = value

    def __getitem__(self, key):
        return self._array[key]

    def __iter__(self):
        """Return the components of the vector as an iterator.

        Examples::

            >>> p = Vector(2, 3, 4)
            >>> print([ c for c in p])
            [2, 3, 4]

        """
        for k in self._array:
            yield k

    def __eq__(self, other):
        if hasattr(other, '_array') and self._array.shape == other._array.shape:
            return np.all(np.absolute(self._array - other._array) < EPSILON)
        return False

    def __neq__(self, other):
        if hasattr(other, '_array') and self._array.shape == other._array.shape:
            return not np.all(np.absolute(self._array - other._array) < EPSILON)
        return True

    def __repr__(self):
        class_name = self.__class__.__name__
        fvalues = (class_name, self.x, self.y, self.z)
        return "{}({:.2f}, {:.2f}, {:.2f})".format(*fvalues)

    __str__ = __repr__



#
# Part of p5: A Python package based on Processing
# Copyright (C) 2017-2019 Abhik Pal
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

from . import p5
from . import primitives
from .shape import PShape

from ..pmath import curves

shape_kind = None
vertices = [] # stores the vertex coordinates
vertices_types = [] # stores the type of vertex. Eg: bezier, curve, etc
contour_vertices = []
contour_vertices_types = []
is_bezier = False
is_curve = False
is_quadratic = False
is_contour = False


__all__ = ['begin_shape', 'end_shape', 'begin_contour', 'end_contour',
		   'curve_vertex', 'bezier_vertex', 'quadratic_vertex', 'vertex']

def begin_shape(kind=None):
	"""
	Begin shape drawing.  This is a helpful way of generating custom shapes quickly.

	:param kind: POINTS, LINES, TRIANGLES, TRIANGLE_FAN TRIANGLE_STRIP, QUADS, or QUAD_STRIP
	:type kind: str
	"""
	global shape_kind, vertices, contour_vertices, vertices_types, contour_vertices_types, is_contour
	if (
		kind == "POINTS" or
		kind == "LINES" or
		kind == "TRIANGLES" or
		kind == "TRIANGLE_FAN" or
		kind == "TRIANGLE_STRIP" or 
		kind == "QUADS" or 
		kind == "QUAD_STRIP"
		):
		shape_kind = kind
	else:
		shape_kind = None

	is_contour = False
	vertices = []
	contour_vertices = []
	vertices_types = []
	contour_vertices_types = []

def curve_vertex(x, y, z=0):
	"""
	Specifies vertex coordinates for curves. The first 
	and last points in a series of curveVertex() lines 
	will be used to guide the beginning and end of a the 
	curve. A minimum of four points is required to draw a 
	tiny curve between the second and third points. Adding 
	a fifth point with curveVertex() will draw the curve 
	between the second, third, and fourth points. The 
	curveVertex() function is an implementation of 
	Catmull-Rom splines.

	:param x: x-coordinate of the vertex
	:type x: float

	:param y: y-coordinate of the vertex
	:type y: float

	:param z: z-coordinate of the vertex
	:type z: float
	"""

	global is_curve, vertices, contour_vertices, is_contour
	global vertices_types, contour_vertices_types
	is_curve = True

	if p5.mode == "3D":
		return
	else:
		if is_contour:
			contour_vertices.append((x, y, z))
			contour_vertices_types.append(2)
		else:
			vertices.append((x, y, z)) # False attribute if the vertex is 
			vertices_types.append(2)

def bezier_vertex(x2, y2, x3, y3, x4, y4):
	"""
	Specifies vertex coordinates for Bezier curves

	:param x2: x-coordinate of the first control point
	:type x2: float

	:param y2: y-coordinate of the first control point
	:type y2: float

	:param x3: x-coordinate of the second control point
	:type x3: float

	:param y3: y-coordinate of the second control point
	:type y3: float

	:param x4: x-coordinate of the anchor point
	:type x4: float

	:param y4: y-coordinate of the anchor point
	:type y4: float
	"""
	global is_bezier, vertices, contour_vertices
	global vertices_types, contour_vertices_types, is_contour
	is_bezier = True

	if p5.mode == "3D":
		return
	else:
		if is_contour:
			contour_vertices.append((x2, y2, x3, y3, x4, y4))
			contour_vertices_types.append(3)
		else:
			vertices.append((x2, y2, x3, y3, x4, y4)) 
			vertices_types.append(3)

def quadratic_vertex(cx, cy, x3, y3):
	"""
	Specifies vertex coordinates for quadratic Bezier curves

	:param cx: x-coordinate of the control point
	:type cx: float

	:param cy: y-coordinate of the control point
	:type cy: float

	:param x3: x-coordinate of the anchor point
	:type x3: float

	:param y3: y-coordinate of the anchor point
	:type y3: float

	"""
	global is_quadratic, vertices, contour_vertices
	global vertices_types, contour_vertices_types, is_contour
	is_quadratic = True

	if p5.mode == "3D":
		return
	else:
		if is_contour:
			contour_vertices.append((cx, cy, x3, y3))
			contour_vertices_types.append(4)
		else:
			vertices.append((cx, cy, x3, y3)) 
			vertices_types.append(3)

def vertex(x, y, z=0):
	"""
	All shapes are constructed by connecting a series of 
	vertices. vertex() is used to specify the vertex 
	coordinates for points, lines, triangles, quads, 
	and polygons. It is used exclusively within the 
	beginShape() and endShape() functions.

	:param x: x-coordinate of the vertex
	:type x: float

	:param y: y-coordinate of the vertex
	:type y: float

	:param z: z-coordinate of the vertex
	:type z: float
	"""
	global vertices, contour_vertices, vertices_types, contour_vertices_types
	global is_contour
	if p5.mode == "3D":
		return
	else:
		if is_contour:
			contour_vertices.append((x, y))
			contour_vertices_types.append(1)
		else:
			vertices.append((x, y))
			vertices_types.append(1)

def begin_contour():
	"""
	Use the beginContour() and endContour() functions 
	to create negative shapes within shapes such as 
	the center of the letter 'O'. beginContour() begins 
	recording vertices for the shape and endContour() stops 
	recording. The vertices that define a negative shape must 
	"wind" in the opposite direction from the exterior shape. 
	First draw vertices for the exterior clockwise order, then 
	for internal shapes, draw vertices shape in counter-clockwise. 

	"""
	global is_contour, contour_vertices, contour_vertices_types
	is_contour = True
	contour_vertices = []
	contour_vertices_types = []

def end_contour():
	global is_contour
	is_contour = False

def get_curve_vertices(verts):
	if len(verts) == 0:
		return []

	s = 1 - curves.curve_tightness_amount
	shape_vertices = [(verts[1][0], verts[1][1], 0.0)]
	steps = curves.curve_resolution

	for i in range(1, len(verts) - 2):
		v = verts[i]
		start = (shape_vertices[len(shape_vertices) - 1][0], shape_vertices[len(shape_vertices) - 1][1])
		c1 = [
			v[0] + (s * verts[i + 1][0] - s * verts[i - 1][0]) / 6,
			v[1] + (s * verts[i + 1][1] - s * verts[i - 1][1]) / 6
		]
		c2 = [
			verts[i + 1][0] +
			(s * verts[i][0] - s * verts[i + 2][0]) / 6,
			verts[i + 1][1] + (s * verts[i][1] - s * verts[i + 2][1]) / 6
		]
		stop = [verts[i + 1][0], verts[i + 1][1]]
		
		for i in range(steps + 1):
			t = i / steps
			p = curves.bezier_point(start, c1, c2, stop, t)
			shape_vertices.append(p[:3])

	return shape_vertices

def get_bezier_vertices(verts, vert_types):
	if len(verts) == 0:
		return []

	shape_vertices = []
	steps = curves.curve_resolution
	for i in range(len(verts)):
		if vert_types[i] == 1 or vert_types[i] == 2:
			shape_vertices.append((verts[i][0], verts[i][1], 0.0))
		else:
			start = (shape_vertices[len(shape_vertices) - 1][0], shape_vertices[len(shape_vertices) - 1][1])
			c1 = [verts[i][0], verts[i][1]]
			c2 = [verts[i][2], verts[i][3]]
			stop = [verts[i][4], verts[i][5]]

			for i in range(steps + 1):
					t = i / steps
					p = curves.bezier_point(start, c1, c2, stop, t)
					shape_vertices.append(p[:3])

	return shape_vertices

def get_quadratic_vertices(verts, vert_types):
	if len(verts) == 0:
		return []
	
	shape_vertices = []
	steps = curves.curve_resolution
	for i in range(len(verts)):
		if vert_types[i] == 1 or vert_types[i] == 2:
			shape_vertices.append((verts[i][0], verts[i][1], 0.0))
		else:
			start = (shape_vertices[len(shape_vertices) - 1][0], shape_vertices[len(shape_vertices) - 1][1])
			control = [verts[i][0], verts[i][1]]
			stop = [verts[i][2], verts[i][3]]

			for i in range(steps + 1):
				t = i / steps
				p = curves.quadratic_point(start, control, stop, t)
				shape_vertices.append(p[:3])

	return shape_vertices

@primitives._draw_on_return
def end_shape(mode=""):
	"""
	The endShape() function is the companion to beginShape()
	and may only be called after beginShape(). When endshape()
	is called, all of image data defined since the previous call
	to beginShape() is rendered.

	:param mode: use CLOSE to close the shape
	:type mode: str

	"""
	global vertices, vertices_types, is_bezier, is_curve, is_quadratic, is_contour, shape_kind

	if len(vertices) == 0:
		return

	if (not p5.renderer.stroke_enabled) and (not p5.renderer.fill_enabled):
		return

	close_shape = mode == "CLOSE"

	# if the shape is closed, the first element is also the last element
	if close_shape:
		attribs = "closed"
		if not is_contour:
			vertices.append(vertices[0])		
	else:
		attribs = "open"

	shape = PShape([(0,0)], children=[])
	if is_curve and (shape_kind == "POLYGON" or shape_kind == None):
		if len(vertices) > 3:
			shape.add_child(PShape(get_curve_vertices(vertices), contour=get_curve_vertices(contour_vertices), attribs=attribs))
	elif is_bezier and (shape_kind == "POLYGON" or shape_kind == None):
		shape.add_child(PShape(get_bezier_vertices(vertices, vertices_types), contour=get_bezier_vertices(contour_vertices, contour_vertices_types), attribs=attribs))
	elif is_quadratic and (shape_kind == "POLYGON" or shape_kind == None):
		shape.add_child(PShape(get_quadratic_vertices(vertices, vertices_types), contour=get_bezier_vertices(contour_vertices, contour_vertices_types), attribs=attribs))
	else:
		if shape_kind == "POINTS":
			shape.add_child(PShape(vertices, attribs='point'))
		elif shape_kind == "LINES":
			if len(vertices) < 2:
				raise ValueError("Insufficient number of vertices %s" % (len(vertices)))
			else:
				for i in range(0, len(vertices), 2):
					shape.add_child(PShape([vertices[i], vertices[i + 1]], attribs='path'))
		elif shape_kind == "TRIANGLES":
			if len(vertices) < 3:
				raise ValueError("Insufficient number of vertices %s" % (len(vertices)))
			else:
				for i in range(0, len(vertices) - 2, 3):
					shape.add_child(PShape([vertices[i], vertices[i + 1], vertices[i + 2]]))
		elif shape_kind == "TRIANGLE_STRIP":
			if len(vertices) < 3:
				raise ValueError("Insufficient number of vertices %s" % (len(vertices)))
			else:
				for i in range(0, len(vertices) - 2, 1):
					shape.add_child(PShape([vertices[i], vertices[i + 1], vertices[i + 2]]))
		elif shape_kind == "TRIANGLE_FAN":
			if len(vertices) < 3:
				raise ValueError("Insufficient number of vertices %s" % (len(vertices)))
			else:
				for i in range(1, len(vertices) - 1):
					shape.add_child(PShape([vertices[0], vertices[i], vertices[i + 1]]))
		elif shape_kind == "QUADS":
			if len(vertices) < 4:
				raise ValueError("Insufficient number of vertices %s" % (len(vertices)))
			else:
				for i in range(0, len(vertices) - 3, 4):
					shape.add_child(PShape([vertices[i], vertices[i + 1], vertices[i + 2], vertices[i + 3]]))
		elif shape_kind == "QUAD_STRIP":
			if len(vertices) < 4:
				raise ValueError("Insufficient number of vertices %s" % (len(vertices)))
			else:
				for i in range(0, len(vertices) - 2, 2):
					shape.add_child(PShape([vertices[i], vertices[i + 1], vertices[i + 3], vertices[i + 2]]))
		else:
			shape.add_child(PShape(vertices, contour=contour_vertices, attribs=attribs))

	is_bezier = False
	is_curve = False
	is_quadratic = False
	is_contour = False
	
	return shape



#
# Part of p5: A Python package based on Processing
# Copyright (C) 2017-2019 Abhik Pal
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
import contextlib
import functools

import builtins

import numpy as np
from PIL import Image
from PIL import ImageFilter
from PIL import ImageChops
from PIL import ImageOps
from vispy import gloo
from . import p5

from . import color
from ..pmath import constrain
from ..pmath.utils import _is_numeric
from .structure import push_style

__all__ = ['PImage', 'image', 'load_image', 'image_mode',
           'load_pixels']

_image_mode = 'corner'

def _ensure_loaded(func):
    """Reloads the image if required before calling the function.

    """
    @functools.wraps(func)
    def rfunc(instance, *args, **kwargs):
        if instance._img_data is None or instance._reload:
            instance._load()
        return func(instance, *args, **kwargs)
    return rfunc

@contextlib.contextmanager
def _restore_color_mode():
    old_mode = color.color_parse_mode
    old_range = color.color_range
    color.color_mode('RGB', 255, 255, 255, 255)

    yield

    color.color_mode(old_mode, *old_range)

class PImage:
    """Image class for p5.

    Note that the image "behaves" like a 2-D list and hence, doesn't
    expose special methods for copying / pasting / cropping. All of
    these operations can be done by using appropriate indexing into
    the array. See the :meth:`p5.PImage.__getitem__` and
    :meth:`p5.PImage.__setitem__` methods for details.

    :param width: width of the image.
    :type width: int

    :param height: height of the image.
    :type height:

    :param fmt: color format to use for the image. Should be one of
        {'RGB', 'RGBA', 'ALPHA'}. Defaults to 'RGBA'

    :type fmt: str

    """
    def __init__(self, width, height, fmt='RGBA'):
        self._width = width
        self._height = height

        format_map = {
            'rgb': 'RGB',
            'rgba': 'RGBA',
            'alpha': 'L',
        }

        self._reload = False
        self._img = None
        self._img_format = format_map[fmt.lower()]
        self._img_texture = None
        self._img_data = None

    @property
    @_ensure_loaded
    def width(self):
        """The width of the image

        :rtype: int
        """
        return self._width

    @width.setter
    def width(self, new_width):
        self.size = (new_width, self._height)

    @property
    @_ensure_loaded
    def height(self):
        """The height of the image

        :rtype: int 
        """
        return self._height

    @height.setter
    def height(self, new_height):
        self.size = (self._width, new_height)

    @property
    @_ensure_loaded
    def size(self):
        """The size of the image

        :rtype: (int, int) tuple
        """
        return self._size

    @size.setter
    def size(self, new_size):
        self._img = self._img.resize(new_size)
        self._reload = True

    @property
    @_ensure_loaded
    def aspect_ratio(self):
        """Return the aspect ratio of the image.

        :rtype: float | int
        """
        return self._width / self._height

    @property
    @_ensure_loaded
    def _texture(self):
        if self._img_texture is None:
            texdata = self._data.astype(np.float32) / 255.0
            self._img_texture = gloo.Texture2D(texdata, interpolation='linear')
        return self._img_texture

    @property
    @_ensure_loaded
    def _data(self):
        return self._img_data

    def _load(self):
        if self._img is None:
            self._img = Image.new(self._img_format, (self._width, self._height))

        width, height = self._img.size
        self._width = width
        self._height = height
        self._size = (width, height)

        data = np.array(self._img.getdata(), dtype=np.uint8)

        self._channels = len(self._img.getbands())

        self._img_data = data.reshape(height, width, self._channels)
        self._img_texture = None
        self._reload = False

    @_ensure_loaded
    def _get_pixel(self, key):
        """Return the pixel color at the given positions.

        :param key: An (x, y) tuple specifying the pixel location.
        :type key: tuple

        :retuns: the Color of the given pixel in the image.
        :rtype: p5.Color

        :raises KeyError: When the pixel location is invalid.

        """
        px = int(key[0])
        py = int(key[1])

        if px >= self.width or py >= self.height:
            raise KeyError("Invalid pixel coordinates {}.".format(key))

        with _restore_color_mode():
            col = color.Color(*self._img.getpixel((px, py)))

        return col

    @_ensure_loaded
    def _get_patch(self, key):
        """Return the patch (a sub-image) specified by the given key.

        :param key: a tuple index-ing into a valid 2D array.
        :type key: tuple

        :returns: the patch specified by the given key
        :rtype: p5.PImage

        """
        xidx, yidx = key
        if self._channels == 1:
            patch_data = self._img_data[yidx, xidx]
        else:
            patch_data = self._img_data[yidx, xidx, :]

        patch_width = patch_data.shape[1]
        patch_height = patch_data.shape[0]

        patch_img = Image.fromarray(patch_data, self._img.mode)

        patch = PImage(patch_width, patch_height)
        patch._img_format = self._img_format
        patch._img = patch_img
        return patch

    @_ensure_loaded
    def __getitem__(self, key):
        """Return the color of the indexed pixel or the requested sub-region

        Note :: when the specified `key` denotes a single pixel, the
            color of that pixel is returned. Else, a new PImage
            (constructed using the slice specified by `key`). Note
            that this causes the internal buffer data to be reloaded
            (when the image is in an "unclean" state) and hence, many
            such operations can potentially slow things down.

        :returns: a sub-image or a the pixel color
        :rtype: p5.Color | p5.PImage

        :raises KeyError: When `key` is invalid.

        """
        if len(key) != 2:
            raise KeyError("Invalid image index")

        if _is_numeric(key[0]) and _is_numeric(key[1]):
            return self._get_pixel(key)

        return self._get_patch(key)

    def _set_pixel(self, key, point):
        # since both posx and posy are numeric, we are only pasting in
        # a "single" color. Hence, whenever patch is a PImage, we
        # require it to be exactly one pixel by one pixel large

        if isinstance(point, PImage):
            if point.size != (1, 1):
                raise AttributeError("Incompatible image dimensions")
            # ...and extract the color of that pixel
            point_color = point[0, 0]

        # if the patch is a color, we don't do much
        elif isinstance(point, color.Color):
            point_color = point

        # otherwise, we try to parse the given patch as a color (if
        # `value` is invalid, the parsing will fail).
        else:
            # the try-catch just takes care of the situation when
            # patch is just a single (non-iterable) value. Then we
            # don't unpack it.
            try:
                point_color = color.Color(*point)
            except TypeError as te:
                point_color = color.Color(point)

        # finally, write the color value to the image based on the
        # number of channels.
        with _restore_color_mode():
            red = int(point_color.red)
            green = int(point_color.green)
            blue = int(point_color.blue)
            alpha = int(point_color.alpha)

            if self._channels == 1:
                pixel_value = int(point_color.gray)
            elif self._channels == 3:
                pixel_value = (red, green, blue)
            elif self._channels == 4:
                pixel_value = (red, green, blue, alpha)
            else:
                raise ValueError("Image has unexpected number of channels")

        pixel_position = int(key[0]), int(key[1])
        self._img.putpixel(pixel_position, pixel_value)

    @_ensure_loaded
    def _set_patch(self, key, patch):
        """Paste the given patch in the image.

        """
        # we first ensure that both the source patch and the target
        # image (self) have the same color modes, if not, convert the
        # target.
        target_mode = self._img.mode
        source_mode = patch._img.mode

        kx, ky = key

        if target_mode != source_mode:
            patch._img.convert(target_mode)

        if self._channels == 1:
            self._img_data[ky, kx] = patch._data[:, :]
        else:
            self._img_data[ky, kx, :] = patch._data[:, :, :]

        self._img = Image.fromarray(self._img_data, self._img.mode)

    def __setitem__(self, key, patch):
        """Paste the given `patch` into the current image.

        """
        if len(key) != 2:
            raise KeyError("Invalid image index")
        if _is_numeric(key[0]) and _is_numeric(key[1]):
            self._set_pixel(key, patch)
        else:
            self._set_patch(key, patch)
        self._reload = True

    def load_pixels(self):
        """Load internal pixel data for the image.
        
        By default image data is only loaded lazily, i.e., right
        before displaying an image on the screen. Use this method to
        manually load the internal image data.

        """
        self._load()

    def mask(self, image):
        raise NotImplementedError

    def filter(self, kind, param=None):
        """Filter the image.

        :param kind: The kind of filter to use on the image. Should be
            one of { 'threshold', 'gray', 'opaque', 'invert',
            'posterize', 'blur', 'erode', 'dilate', }

        :type kind: str

        :param param: optional parameter for the filter in use
            (defaults to None). Only required for 'threshold' (the
            threshold to use, param should be a value between 0 and 1;
            defaults to 0.5), 'posterize' (limiting value for each
            channel should be between 2 and 255), and 'blur' (gaussian
            blur radius, defaults to 1.0).

        :type param: int | float | None

        """
        filter_name = kind.lower()
        if param is None:
            default_values = {
                'threshold': 0.5,
                'blur': 1.0,
                'gaussian_blur': 1.0,
                'box_blur': 1.0,
                'posterize': 1,
                'opacity': 0.5,
            }
            param = default_values.get(filter_name, None)

        if filter_name in ['blur', 'gaussian_blur']:
            fim = self._img.filter(ImageFilter.GaussianBlur(radius=param))
            self._img = fim
        elif filter_name == 'box_blur':
            self._img = self._img.filter(ImageFilter.BoxBlur(param))
        elif filter_name in ['gray', 'grey', 'grayscale']:
            self._img = ImageOps.grayscale(self._img)
        elif filter_name == 'opaque':
            self._img.putalpha(255)
        elif filter_name == 'opacity':
            self._img.putalpha(int(param * 255))
        elif filter_name == 'invert':
            self._img = ImageOps.invert(self._img)
        elif filter_name == 'posterize':
            nbits = 0
            while int(param) != 0:
                param = param >> 1
                nbits = nbits + 1
            nbits = constrain(nbits, 1, 8)
            self._img = ImageOps.posterize(self._img, nbits)
        elif filter_name == 'threshold':
            dat = np.asarray(ImageOps.grayscale(self._img)).copy()
            dat[dat < int(128 * param)] = 0
            dat[dat >= int(128 * param)] = 255
            self._img = Image.fromarray(dat)
        elif filter_name in ['erode', 'dilate']:
            raise NotImplementedError
        else:
            raise ValueError("Unknown filter")

        self._reload = True

    def blend(self, other, mode):
        """Blend the specified image using the given blend mode.

        :param other: The image to be blended to the current image.
        :type other: p5.PImage

        :param mode: Blending mode to use. Should be one of { 'BLEND',
            'ADD', 'SUBTRACT', 'LIGHTEST', 'DARKEST', 'MULTIPLY',
            'SCREEN',}
        :type mode: str

        :raises AssertionError: When the dimensions of img do not
            match the dimensions of the current image.

        :raises KeyError: When the blend mode is invalid.

        """
        mode = mode.lower()
        assert self.size == other.size, "Images are of different sizes!"

        if self._img.mode != 'RGBA':
            self._img = self._img.convert('RGBA')
            self._reload = True

        if other._img.mode != 'RGBA':
            other_img = other._img.convert('RGBA')
        else:
            other_img = other._img

        # todo: implement missing filters -- abhikpal (2018-08-14)
        if mode == 'blend':
            self._img = ImageChops.composite(self._img, other_img, self._img)
        elif mode == 'add':
            self._img = ImageChops.add(self._img, other_img)
        elif mode == 'subtract':
            self._img = ImageChops.subtract(self._img, other_img)
        elif mode == 'lightest':
            self._img = ImageChops.lighter(self._img, other_img)
        elif mode == 'darkest':
            self._img = ImageChops.darker(self._img, other_img)
        elif mode == 'difference':
            raise NotImplementedError
        elif mode == 'exclusion':
            raise NotImplementedError
        elif mode == 'multiply':
            self._img = ImageChops.multiply(self._img, other_img)
        elif mode == 'screen':
            self._img = ImageChops.screen(self._img, other_img)
        elif mode == 'overlay':
            raise NotImplementedError
        elif mode == 'hard_light':
            raise NotImplementedError
        elif mode == 'soft_light':
            raise NotImplementedError
        elif mode == 'dodge':
            raise NotImplementedError
        elif mode == 'burn':
            raise NotImplementedError
        else:
            raise KeyError("'{}' blend mode not found".format(mode.upper()))

        self._reload = True
        return self

    @_ensure_loaded
    def save(self, file_name):
        """Save the image into a file

        :param file_name: Filename to save the image as.
        :type file_name: str

        """
        self._img.save(file_name)

def image(img, location, size=None):
    """Draw an image to the display window.

    Images must be in the same folder as the sketch (or the image path
    should be explicitly mentioned). The color of an image may be
    modified with the :meth:`p5.tint` function. 

    :param img: the image to be displayed.
    :type img: p5.Image

    :param location: location of the image on the screen (depending on the
        current image mode, 'corner', 'center', 'corners', this could
        represent the coordinate of the top-left corner, center,
        top-left corner respectively.)
    :type location: tuple | list | np.ndarray | p5.Vector

    :param size: target size of the image or the bottom-right image
        corner when the image mode is set to 'corners'. By default,
        the value is set according to the current image size.

    :type size: tuple | list

    """
    if size is None:
        size = img.size

    lx, ly = location
    sx, sy = size

    if _image_mode == 'center':
        lx = int(lx - (sx / 2))
        ly = int(ly - (sy / 2))

    if _image_mode == 'corners':
        sx = sx - lx
        sy = sy - ly

    p5.renderer.render_image(img, (lx, ly), (sx, sy))

def image_mode(mode):
    """Modify the locaton from which the images are drawn.

    Modifies the location from which images are drawn by changing the
    way in which parameters given to :meth:`p5.image` are intepreted.

    The default mode is ``image_mode('corner')``, which interprets the
    second parameter of ``image()`` as the upper-left corner of the
    image. If an additional parameter is specified, it is used to set
    the image's width and height.

    ``image_mode('corners')`` interprets the first parameter of
    ``image()`` as the location of one corner, and the second
    parameter as the opposite corner.

    ``image_mode('center')`` interprets the first parameter of
    ``image()`` as the image's center point. If an additional
    parameter is specified, it is used to set the width and height of
    the image.

    :param mode: should be one of ``{'corner', 'center', 'corners'}``
    :type mode: str

    :raises ValueError: When the given image mode is not understood.
        Check for typoes.

    """
    global _image_mode

    if mode.lower() not in ['corner', 'center', 'corners']:
        raise ValueError("Unknown image mode!")
    _image_mode = mode.lower()

def load_image(filename):
    """Load an image from the given filename.

    Loads an image into a variable of type PImage. Four types of
    images may be loaded. 

    In most cases, load all images in setup() or outside the draw()
    call to preload them at the start of the program. Loading images
    inside draw() will reduce the speed of a program. 

    :param filename: Filename (or path)of the given image. The
        file-extennsion is automatically inferred.
    :type filename: str

    :returns: An :class:`p5.PImage` instance with the given image data
    :rtype: :class:`p5.PImage`

    """
    # todo: add support for loading images from URLs -- abhikpal
    # (2018-08-14)
    img = Image.open(filename)
    w, h = img.size
    pimg = PImage(w, h)
    pimg._img = img
    return pimg

@contextlib.contextmanager
def load_pixels():
    """Load a snapshot of the display window into the ``pixels`` Image.
    
    This context manager loads data into the global ``pixels`` Image.
    Once the program execution leaves the context manager, all changes
    to the image are written to the main display.

    """
    pixels = PImage(builtins.width, builtins.height, 'RGB')
    #sketch.renderer.flush_geometry()
    pixel_data = p5.renderer.fbuffer.read(mode='color', alpha=False)

    pixels._img = Image.fromarray(pixel_data)
    builtins.pixels = pixels

    pixels._load()

    yield

    with push_style():
        image_mode('corner')
        p5.renderer.tint_enabled = False
        image(builtins.pixels, (0, 0))

    builtins.pixels = None



#
# Part of p5: A Python package based on Processing
# Copyright (C) 2017-2019 Abhik Pal
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import numpy as np
import math
from ..pmath import matrix

import builtins

from vispy import gloo
from vispy.gloo import FrameBuffer
from vispy.gloo import IndexBuffer
from vispy.gloo import Program
from vispy.gloo import Texture2D
from vispy.gloo import VertexBuffer

from contextlib import contextmanager

from ..core.constants import *

from .shaders2d import src_default
from .shaders2d import src_fbuffer
from .shaders2d import src_texture
from .shaders2d import src_line

class Renderer2D:
	def __init__(self):
		self.default_prog = None
		self.fbuffer_prog = None
		self.texture_prog = None
		self.line_prog = None

		self.fbuffer = None
		self.fbuffer_tex_front = None
		self.fbuffer_tex_back = None

		self.vertex_buffer = None
		self.index_buffer = None

		## Renderer Globals: USEFUL CONSTANTS
		self.COLOR_WHITE = (1, 1, 1, 1)
		self.COLOR_BLACK = (0, 0, 0, 1)
		self.COLOR_DEFAULT_BG = (0.8, 0.8, 0.8, 1.0)

		## Renderer Globals: STYLE/MATERIAL PROPERTIES
		##
		self.background_color = self.COLOR_DEFAULT_BG

		self.fill_color = self.COLOR_WHITE
		self.fill_enabled = True

		self.stroke_color = self.COLOR_BLACK
		self.stroke_enabled = True

		self.tint_color = self.COLOR_BLACK
		self.tint_enabled = False

		## Renderer Globals: Curves
		self.stroke_weight = 1
		self.stroke_cap = 2
		self.stroke_join = 0

		## Renderer Globals
		## VIEW MATRICES, ETC
		##
		self.viewport = None
		self.texture_viewport = None
		self.transform_matrix = np.identity(4)
		self.modelview_matrix = np.identity(4)
		self.projection_matrix = np.identity(4)

		## Renderer Globals: RENDERING
		self.draw_queue = []

	def initialize_renderer(self):
		self.fbuffer = FrameBuffer()

		vertices = np.array([[-1.0, -1.0],
							 [+1.0, -1.0],
							 [-1.0, +1.0],
							 [+1.0, +1.0]],
							np.float32)
		texcoords = np.array([[0.0, 0.0],
							  [1.0, 0.0],
							  [0.0, 1.0],
							  [1.0, 1.0]],
							 dtype=np.float32)

		self.fbuf_vertices = VertexBuffer(data=vertices)
		self.fbuf_texcoords = VertexBuffer(data=texcoords)

		self.fbuffer_prog = Program(src_fbuffer.vert, src_fbuffer.frag)
		self.fbuffer_prog['texcoord'] = self.fbuf_texcoords
		self.fbuffer_prog['position'] = self.fbuf_vertices

		self.vertex_buffer = VertexBuffer()
		self.index_buffer = IndexBuffer()

		self.default_prog = Program(src_default.vert, src_default.frag)
		self.texture_prog = Program(src_texture.vert, src_texture.frag)
		self.texture_prog['texcoord'] = self.fbuf_texcoords

		self.reset_view()

	def reset_view(self):
		self.viewport = (
			0,
			0,
			int(builtins.width * builtins.pixel_x_density),
			int(builtins.height * builtins.pixel_y_density),
		)
		self.texture_viewport = (
			0,
			0,
			builtins.width,
			builtins.height,
		)

		gloo.set_viewport(*self.viewport)

		cz = (builtins.height / 2) / math.tan(math.radians(30))
		self.projection_matrix = matrix.perspective_matrix(
			math.radians(60),
			builtins.width / builtins.height,
			0.1 * cz,
			10 * cz
		)
		self.modelview_matrix = matrix.translation_matrix(-builtins.width / 2, \
													 builtins.height / 2, \
													 -cz)
		self.modelview_matrix = self.modelview_matrix.dot(matrix.scale_transform(1, -1, 1))

		self.transform_matrix = np.identity(4)

		self.default_prog['modelview'] = self.modelview_matrix.T.flatten()
		self.default_prog['projection'] = self.projection_matrix.T.flatten()

		self.texture_prog['modelview'] = self.modelview_matrix.T.flatten()
		self.texture_prog['projection'] = self.projection_matrix.T.flatten()

		self.line_prog = Program(src_line.vert, src_line.frag)

		self.line_prog['modelview'] = self.modelview_matrix.T.flatten()
		self.line_prog['projection'] = self.projection_matrix.T.flatten()
		self.line_prog["height"] = builtins.height

		self.fbuffer_tex_front = Texture2D((builtins.height, builtins.width, 3))
		self.fbuffer_tex_back = Texture2D((builtins.height, builtins.width, 3))

		for buf in [self.fbuffer_tex_front, self.fbuffer_tex_back]:
			self.fbuffer.color_buffer = buf
			with self.fbuffer:
				self.clear()

	def clear(self, color=True, depth=True):
		"""Clear the renderer background."""
		gloo.set_state(clear_color=self.background_color)
		gloo.clear(color=color, depth=depth)

	def _comm_toggles(self, state=True):
		gloo.set_state(blend=state)
		gloo.set_state(depth_test=state)

		if state:
			gloo.set_state(blend_func=('src_alpha', 'one_minus_src_alpha'))
			gloo.set_state(depth_func='lequal')

	@contextmanager
	def draw_loop(self):
		"""The main draw loop context manager.
		"""

		self.transform_matrix = np.identity(4)

		self.default_prog['modelview'] = self.modelview_matrix.T.flatten()
		self.default_prog['projection'] = self.projection_matrix.T.flatten()

		self.fbuffer.color_buffer = self.fbuffer_tex_back

		with self.fbuffer:
			gloo.set_viewport(*self.texture_viewport)
			self._comm_toggles()
			self.fbuffer_prog['texture'] = self.fbuffer_tex_front
			self.fbuffer_prog.draw('triangle_strip')

			yield

			self.flush_geometry()
			self.transform_matrix = np.identity(4)

		gloo.set_viewport(*self.viewport)
		self._comm_toggles(False)
		self.clear()
		self.fbuffer_prog['texture'] = self.fbuffer_tex_back
		self.fbuffer_prog.draw('triangle_strip')

		self.fbuffer_tex_front, self.fbuffer_tex_back = self.fbuffer_tex_back, self.fbuffer_tex_front


	def _transform_vertices(self, vertices, local_matrix, global_matrix):
		return np.dot(np.dot(vertices, local_matrix.T), global_matrix.T)[:, :3]

	def render(self, shape):
		vertices = shape._draw_vertices
		n, _ = vertices.shape
		tverts = self._transform_vertices(
			np.hstack([vertices, np.zeros((n, 1)), np.ones((n, 1))]),
			shape._matrix,
			self.transform_matrix)
		
		fill = shape.fill.normalized if shape.fill else None
		stroke = shape.stroke.normalized if shape.stroke else None
		stroke_weight = shape.stroke_weight
		stroke_cap = shape.stroke_cap
		stroke_join = shape.stroke_join

		edges = shape._draw_edges
		faces = shape._draw_faces
		
		if edges is None:
			print(vertices)
			print("whale")
			exit()

		if 'open' in shape.attribs:
			self.add_to_draw_queue('poly', tverts, edges[:-1], faces, fill, stroke, 
					stroke_weight, stroke_cap, stroke_join)
		else:
			self.add_to_draw_queue(shape.kind, tverts, edges, faces, fill, stroke, 
					stroke_weight, stroke_cap, stroke_join)

	def add_to_draw_queue(self, stype, vertices, edges, faces, fill=None, stroke=None,
			stroke_weight=None, stroke_cap=None, stroke_join=None):
		"""Add the given vertex data to the draw queue.

		:param stype: type of shape to be added. Should be one of {'poly',
			'path', 'point'}
		:type stype: str

		:param vertices: (N, 3) array containing the vertices to be drawn.
		:type vertices: np.ndarray

		:param edges: (N, 2) array containing edges as tuples of indices
			into the vertex array. This can be None when not appropriate
			(eg. for points)
		:type edges: None | np.ndarray

		:param faces: (N, 3) array containing faces as tuples of indices
			into the vertex array. For 'point' and 'path' shapes, this can
			be None
		:type faces: np.ndarray

		:param fill: Fill color of the shape as a normalized RGBA tuple.
			When set to `None` the shape doesn't get a fill (default: None)
		:type fill: None | tuple

		:param stroke: Stroke color of the shape as a normalized RGBA
			tuple. When set to `None` the shape doesn't get stroke
			(default: None)
		:type stroke: None | tuple

		"""

		fill_shape = self.fill_enabled and not (fill is None)
		stroke_shape = self.stroke_enabled and not (stroke is None)

		if fill_shape and stype not in ['point', 'path']:
			idx = np.array(faces, dtype=np.uint32).ravel()
			self.draw_queue.append(["triangles", (vertices, idx, fill)])

		if stroke_shape:
			if stype == 'point':
				idx = np.arange(0, len(vertices), dtype=np.uint32)
				self.draw_queue.append(["points", (vertices, idx, stroke)])
			else:
				self.draw_queue.append(["lines", (
					vertices, edges, stroke, stroke_weight, stroke_cap, stroke_join
					)])

	def flush_geometry(self):
		"""Flush all the shape geometry from the draw queue to the GPU.
		"""
		current_shape = None
		current_queue = []
		for index, shape in enumerate(self.draw_queue):
			current_shape = self.draw_queue[index][0]
			current_queue.append(self.draw_queue[index][1])

			if index < len(self.draw_queue) - 1:
				if self.draw_queue[index][0] == self.draw_queue[index + 1][0]:
					continue

			if current_shape == "points" or current_shape == "triangles":
				self.render_default(current_shape, current_queue)
			elif current_shape == "lines":
				self.render_line(current_queue)

			current_queue = []

		self.draw_queue = []

	def render_default(self, draw_type, draw_queue):
		# 1. Get the maximum number of vertices persent in the shapes
		# in the draw queue.
		#
		if len(draw_queue) == 0:
			return

		num_vertices = 0
		for vertices, _, _ in draw_queue:
			num_vertices = num_vertices + len(vertices)

		# 2. Create empty buffers based on the number of vertices.
		#
		data = np.zeros(num_vertices,
						dtype=[('position', np.float32, 3),
							   ('color', np.float32, 4)])

		# 3. Loop through all the shapes in the geometry queue adding
		# it's information to the buffer.
		#
		sidx = 0
		draw_indices = []
		for vertices, idx, color in draw_queue:
			num_shape_verts = len(vertices)

			data['position'][sidx:(sidx + num_shape_verts),] = vertices

			color_array = np.array([color] * num_shape_verts)
			data['color'][sidx:sidx + num_shape_verts, :] = color_array

			draw_indices.append(sidx + idx)

			sidx += num_shape_verts

		self.vertex_buffer.set_data(data)
		self.index_buffer.set_data(np.hstack(draw_indices))

		# 4. Bind the buffer to the shader.
		#
		self.default_prog.bind(self.vertex_buffer)

		# 5. Draw the shape using the proper shape type and get rid of
		# the buffers.
		#
		self.default_prog.draw(draw_type, indices=self.index_buffer)

	def render_line(self, queue):
		'''
		This rendering algorithm works by tesselating the line into
		multiple triangles.

		Reference: https://blog.mapbox.com/drawing-antialiased-lines-with-opengl-8766f34192dc
		'''

		if len(queue) == 0:
			return

		pos = []
		posPrev = []
		posCurr = []
		posNext = []
		markers = []
		side = []

		linewidth = []
		join_type = []
		cap_type = []
		color = []

		for line in queue:
			if len(line[1]) == 0:
				continue

			for segment in line[1]:
				for i in range(len(segment) - 1): # the data is sent to renderer in line segments
					for j in [0, 0, 1, 0, 1, 1]: # all the vertices of triangles
						if i + j - 1 >= 0:
							posPrev.append(line[0][segment[i + j - 1]])
						else:
							posPrev.append(line[0][segment[i + j]])

						if i + j + 1 < len(segment):
							posNext.append(line[0][segment[i + j + 1]])
						else:
							posNext.append(line[0][segment[i + j]])

						posCurr.append(line[0][segment[i + j]])

					markers.extend([1.0, -1.0, -1.0, -1.0, 1.0, -1.0]) # Is the vertex up/below the line segment
					side.extend([1.0, 1.0, -1.0, 1.0, -1.0, -1.0]) # Left or right side of the segment
					pos.extend([line[0][segment[i]]]*6) # Left vertex of each segment
					linewidth.extend([line[3]]*6)
					join_type.extend([line[5]]*6)
					cap_type.extend([line[4]]*6)
					color.extend([line[2]]*6)

		if len(pos) == 0:
			return 

		posPrev = np.array(posPrev, np.float32)
		posCurr = np.array(posCurr, np.float32)
		posNext = np.array(posNext, np.float32)
		markers = np.array(markers, np.float32)
		side = np.array(side, np.float32)
		pos = np.array(pos, np.float32)
		linewidth = np.array(linewidth, np.float32)
		join_type = np.array(join_type, np.float32)
		cap_type = np.array(cap_type, np.float32)
		color = np.array(color, np.float32)

		self.line_prog['pos'] = gloo.VertexBuffer(pos)
		self.line_prog['posPrev'] = gloo.VertexBuffer(posPrev)
		self.line_prog['posCurr'] = gloo.VertexBuffer(posCurr)
		self.line_prog['posNext'] = gloo.VertexBuffer(posNext)
		self.line_prog['marker'] = gloo.VertexBuffer(markers)
		self.line_prog['side'] = gloo.VertexBuffer(side)
		self.line_prog['linewidth'] = gloo.VertexBuffer(linewidth)
		self.line_prog['join_type'] = gloo.VertexBuffer(join_type)
		self.line_prog['cap_type'] = gloo.VertexBuffer(cap_type)
		self.line_prog["color"] = gloo.VertexBuffer(color)

		self.line_prog.draw('triangles')

	def render_image(self, image, location, size):
		"""Render the image.

		:param image: image to be rendered
		:type image: builtins.Image

		:param location: top-left corner of the image
		:type location: tuple | list | builtins.Vector

		:param size: target size of the image to draw.
		:type size: tuple | list | builtins.Vector
		"""
		self.flush_geometry()

		self.texture_prog['fill_color'] = self.tint_color if self.tint_enabled else self.COLOR_WHITE
		self.texture_prog['transform'] = self.transform_matrix.T.flatten()

		x, y = location
		sx, sy = size
		imx, imy = image.size
		data = np.zeros(4,
						dtype=[('position', np.float32, 2),
							   ('texcoord', np.float32, 2)])
		data['texcoord'] = np.array([[0.0, 1.0],
									 [1.0, 1.0],
									 [0.0, 0.0],
									 [1.0, 0.0]],
									dtype=np.float32)
		data['position'] = np.array([[x, y + sy],
									 [x + sx, y + sy],
									 [x, y],
									 [x + sx, y]],
									dtype=np.float32)

		self.texture_prog['texture'] = image._texture
		self.texture_prog.bind(VertexBuffer(data))
		self.texture_prog.draw('triangle_strip')

	def cleanup(self):
		"""Run the clean-up routine for the renderer.

		This method is called when all drawing has been completed and the
		program is about to exit.

		"""
		self.default_prog.delete()
		self.fbuffer_prog.delete()
		self.line_prog.delete()
		self.fbuffer.delete()




#
# Part of p5: A Python package based on Processing
# Copyright (C) 2017-2019 Abhik Pal
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Shape class for p5.
"""

import contextlib
import functools

import numpy as np
import triangle as tr

from .color import Color
from ..pmath import matrix

from . import p5


__all__ = ['PShape']

def _ensure_editable(func):
    """A decorater that ensures that a shape is in 'edit' mode.

    """
    @functools.wraps(func)
    def editable_method(instance, *args, **kwargs):
        if not instance._in_edit_mode:
            raise ValueError('{} only works in edit mode'.format(func.__name__))
        return func(instance, *args, **kwargs)
    return editable_method

def _apply_transform(func):
    """Apply the matrix transformation to the shape.
    """
    @functools.wraps(func)
    def mfunc(instance, *args, **kwargs):
        tmat = func(instance, *args, **kwargs)
        instance._matrix = instance._matrix.dot(tmat)
        return tmat
    return mfunc

def _call_on_children(func):
    """Call the method on all child shapes
    """
    @functools.wraps(func)
    def rfunc(instance, *args, **kwargs):
        rval = func(instance, *args, **kwargs)
        for child in instance.children:
            rfunc(child, *args, **kwargs)
        return rval
    return rfunc

class PShape:
    """Custom shape class for p5.

    :param vertices: List of (polygonal) vertices for the shape.
    :type vertices: list | np.ndarray

    :param fill_color: Fill color of the shape (default: 'auto' i.e.,
        the current renderer fill)
    :type fill_color: 'auto' | None | tuple | p5.Color

    :param stroke_color: Stroke color of the shape (default: 'auto'
        i.e., the current renderer stroke color)
    :type stroke_color: 'auto' | None | tuple | p5.color

    :param visible: toggles shape visibility (default: False)
    :type visible: bool

    :param attribs: space-separated list of attributes that control
        shape drawing. Each attribute should be one of {'point',
        'path', 'open', 'closed'}. (default: 'closed')
    :type attribs: str

    :param children: List of sub-shapes for the current shape
        (default: [])
    :type children: list

    """
    def __init__(self, vertices=[], fill_color='auto',
                 stroke_color='auto', stroke_weight="auto", 
                 stroke_join="auto", stroke_cap="auto", 
                 visible=False, attribs='closed',
                 children=None, contour=None):
        # basic properties of the shape
        self._vertices = np.array([])
        self.contour = contour or np.array([])
        self._edges = None
        self._outline = None
        self._outline_vertices = None

        self.attribs = set(attribs.lower().split())
        self._fill = None
        self._stroke = None
        self._stroke_weight = None
        self._stroke_cap = None
        self._stroke_join = None

        self._matrix = np.identity(4)
        self._transform_matrix = np.identity(4)
        self._transformed_draw_vertices = None

        # a flag to check if the shape is being edited right now.
        self._in_edit_mode = False
        self._vertex_cache = None

        # The triangulation used to render the shapes.
        self._tri = None
        self._tri_required = not ('point' in self.attribs) and \
                             not ('path' in self.attribs)
        self._tri_vertices = None
        self._tri_edges = None
        self._tri_faces = None

        if len(vertices) > 0:
            self.vertices = vertices

        self.fill = fill_color
        self.stroke = stroke_color
        self.stroke_weight = stroke_weight
        self.stroke_cap = stroke_cap
        self.stroke_join = stroke_join

        self.children = children or []
        self.visible = visible

    def _set_color(self, name, value=None):
        color = None

        if isinstance(value, Color):
            color = value
        else:
            if name == 'stroke' and p5.renderer.stroke_enabled:
                color = Color(*p5.renderer.stroke_color,
                              color_mode='RGBA', normed=True)
            if name == 'fill' and p5.renderer.fill_enabled:
                color = Color(*p5.renderer.fill_color,
                              color_mode='RGBA', normed=True)

        if name == 'stroke':
            self._stroke = color
        elif name == 'fill':
            self._fill = color

    @property
    def fill(self):
        return self._fill

    @fill.setter
    def fill(self, new_color):
        self._set_color('fill', new_color)

    @property
    def stroke(self):
        return self._stroke

    @stroke.setter
    def stroke(self, new_color):
        self._set_color('stroke', new_color)

    @property
    def stroke_weight(self):
        return self._stroke_weight

    @stroke_weight.setter
    def stroke_weight(self, stroke):
        if stroke == "auto":
            self._stroke_weight = p5.renderer.stroke_weight
        else:
            self._stroke_weight = stroke

    @property
    def stroke_join(self):
        return self._stroke_join

    @stroke_join.setter
    def stroke_join(self, stroke):
        if stroke == "auto":
            self._stroke_join = p5.renderer.stroke_join
        else:
            self._stroke_join = stroke

    @property
    def stroke_cap(self):
        return self._stroke_cap

    @stroke_cap.setter
    def stroke_cap(self, stroke):
        if stroke == "auto":
            self._stroke_cap = p5.renderer.stroke_cap
        else:
            self._stroke_cap = stroke

    @property
    def kind(self):
        if 'point' in self.attribs:
            return 'point'
        elif 'path' in self.attribs:
            return 'path'
        else:
            return 'poly'

    def _sanitize_vertex_list(self, vertices, tdim=2, sdim=3):
        """Convert all vertices to the given dimensions.
        Removes consecutive duplicates to prevent errors in triangulation.

        :param vertices: List of vertices
        :type vertices: list

        :param tdim: Target dimension for sanitization (default: 3)
        :type tdim: int

        :param sdim: Source dimension for the points (default: 2).
            Whenever sdim > tdim, the last (sdim - tdim) components will
            be discarded.
        :type sdim: int

        :raises ValueError: when the point dimension is between sdim and tdim

        :returns: A sanitized array of vertices.
        :type: np.ndarray

        """
        sanitized = []
        for i in range(len(vertices)):
            if i < len(vertices) - 1:
                if vertices[i] == vertices[i + 1]:
                    continue
            elif i == len(vertices) - 1 and i != 0:
                if vertices[i] == vertices[0]:
                    continue

            v = vertices[i]
            if (len(v) > max(tdim, sdim)) or (len(v) < min(tdim, sdim)):
                raise ValueError("unexpected vertex dimension")

            if tdim > sdim:
                sanitized.append(list(v) + [0] * (tdim - sdim))
            elif tdim < sdim:
                sanitized.append(list(v)[:tdim])
            else:
                sanitized.append(list(v))

        return np.array(sanitized)

    @property
    def vertices(self):
        return self._vertices

    @vertices.setter
    def vertices(self, new_vertices):
        self._vertices = self._sanitize_vertex_list(new_vertices)

        n = len(self._vertices)
        self._outline_vertices = np.hstack([self._vertices, np.zeros((n, 1))])
        self._tri_vertices = None
        self._tri_edges = None
        self._tri_faces = None

    def _compute_poly_edges(self):
        n, _ = self._vertices.shape
        return np.vstack([np.arange(n), (np.arange(n) + 1) % n]).transpose()

    def _compute_outline_edges(self):
        n, _ = self._vertices.shape
        return np.vstack([np.arange(n - 1),
                          (np.arange(n - 1) + 1) % n]).transpose()

    @property
    def edges(self):
        if 'point' in self.attribs:
            return np.array([])

        if self._edges is None:
            n, _ = self._vertices.shape

            if 'point' in self.attribs:
                self._edges = np.array([])
            elif 'path' in self.attribs:
                self._edges = self._compute_outline_edges()
            else:
                self._edges = self._compute_poly_edges()

            if 'open' in self.attribs:
                self._outline = self._compute_outline_edges()
            else:
                self._outline = self._edges

        return self._edges

    def get_interior_point(self, shape_vertices):
        # Returns a random point inside the shape
        if len(shape_vertices) < 2:
            return []

        # Triangulate the shape
        triangulate = tr.triangulate(dict(vertices=shape_vertices), "a5")
        for vertex in triangulate["vertices"]:
            if vertex not in shape_vertices:
                return [vertex]

        return []

    def _retriangulate(self):
        """Triangulate the shape
        """
        if len(self.vertices) < 2:
            self._tri_edges = np.array([])
            self._tri_faces = np.array([])
            self._tri_vertices = self.vertices
            return

        if len(self.contour) > 1:
            n, _ = self.contour.shape
            contour_edges = np.vstack([np.arange(n), (np.arange(n) + 1) % n]).transpose()
            triangulation_vertices = np.vstack([self.vertices, self.contour])
            triangulation_segments = np.vstack([self.edges, contour_edges + len(self.edges)])
            triangulate_parameters = dict(vertices=triangulation_vertices, 
                segments=triangulation_segments, holes=self.get_interior_point(self.contour))

            self._tri = tr.triangulate(triangulate_parameters, "p")
        else:
            triangulate_parameters = dict(vertices=self.vertices, segments=self.edges)
            self._tri = tr.triangulate(triangulate_parameters, "p")

        if "segments" in self._tri.keys():
            self._tri_edges = self._tri["segments"]
        else:
            self._tri_edges = self.edges

        self._tri_faces = self._tri["triangles"]
        self._tri_vertices = self._tri["vertices"]

    @property
    def _draw_outline_vertices(self):
        if 'open' in self.attribs:
            return self._draw_vertices
        return self.vertices

    @property
    def _draw_outline_edges(self):
        if 'open' in self.attribs:
            return self._outline
        return self._edges

    @property
    def _draw_vertices(self):
        if self._tri_required and (self._tri_vertices is None):
            self._retriangulate()

        if self._tri_required:
            return self._tri_vertices
        return self._vertices

    @property
    def _draw_edges(self):
        if self._tri_required:
            if self._tri_edges is None:
                self._retriangulate()
            return self._tri_edges
        return self.edges

    @property
    def _draw_faces(self):
        if self._tri_required:
            if self._tri_faces is None:
                self._retriangulate()
            return self._tri_faces

        return np.array([])

    @contextlib.contextmanager
    def edit(self, reset=True):
        """Put the shape in edit mode.

        :param reset: Toggles whether the shape should be "reset"
            during editing. When set to `True` all existing shape
            vertices are cleared. When set to `False` the new vertices
            are appended at the end of the existing vertex list.
            (default: True)
        :type reset: bool

        :raises ValueError: if the shape is already being edited.

        """
        if self._in_edit_mode:
            raise ValueError("Shape is being edited already")

        self._in_edit_mode = True
        if reset:
            self._vertices = np.array([])
        self._vertex_cache = []
        yield
        self.vertices = self._vertex_cache
        self._in_edit_mode = False
        self._edges = None

    @_ensure_editable
    def add_vertex(self, vertex):
        """Add a vertex to the current shape

        :param vertex: The (next) vertex to add to the current shape.
        :type vertex: tuple | list | p5.Vector | np.ndarray

        :raises ValueError:  when the vertex is of the wrong dimension
        """
        self._vertex_cache.append(vertex)

    def update_vertex(self, idx, vertex):
        """Edit an indicidual vertex.

        :param idx: index of the vertex to be edited
        :type idx: int

        :param vertex: The (next) vertex to add to the current shape.
        :type vertex: tuple | list | p5.Vector | np.ndarray

        :raises ValueError:  when the vertex is of the wrong dimension
        """
        if len(vertex) != 2:
            raise ValueError("Wrong vertex dimension")
        self._vertices[idx] =  np.array(vertex)
        self._tri_vertices = None
        self._tri_edges = None
        self._tri_faces = None
        self._edges = None

    def add_child(self, child):
        """Add a child shape to the current shape

        :param child: Child to be added
        :type child: PShape
        """
        self.children.append(child)

    def transform_matrix(self, mat):
        self._transform_matrix = mat

    @property
    def child_count(self):
        """Number of children.

        :returns: The current number of children.
        :rtype: int
        """
        return len(self.children)

    def apply_matrix(self, mat):
        """Apply the given transformation matrix to the shape.

        :param mat: the 4x4 matrix to be applied to the current shape.
        :type mat: (4, 4) np.ndarray

        """
        self._matrix = self._matrix.dot(mat)

    @_call_on_children
    def apply_transform_matrix(self, mat):
        self._matrix = self._matrix.dot(mat)

    @_call_on_children
    def reset_matrix(self):
        """Reset the transformation matrix associated with the shape.

        """
        self._matrix = np.identity(4)

    @_call_on_children
    @_apply_transform
    def translate(self, x, y, z=0):
        """Translate the shape origin to the given location.

        :param x: The displacement amount in the x-direction (controls
            the left/right displacement)
        :type x: int

        :param y: The displacement amount in the y-direction (controls
            the up/down displacement)
        :type y: int

        :param z: The displacement amount in the z-direction (0 by
            default). This controls the displacement away-from/towards
            the screen.
        :type z: int

        :returns: The translation matrix applied to the transform
            matrix.
        :rtype: np.ndarray

        """
        tmat = matrix.translation_matrix(x, y, z)
        return tmat

    @_call_on_children
    @_apply_transform
    def rotate(self, theta, axis=[0, 0, 1]):
        """Rotate the shape by the given angle along the given axis.

        :param theta: The angle by which to rotate (in radians)
        :type theta: float

        :param axis: The axis along which to rotate (defaults to the
            z-axis)
        :type axis: np.ndarray | list

        :returns: The rotation matrix used to apply the
            transformation.
        :rtype: np.ndarray

        """
        axis = np.array(axis[:])
        tmat = matrix.rotation_matrix(axis, theta)
        return tmat

    def rotate_x(self, theta):
        """Rotate the shape along the x axis.

        :param theta: angle by which to rotate (in radians)
        :type theta: float

        :returns: The rotation matrix used to apply the
            transformation.
        :rtype: np.ndarray

        """
        return self.rotate(theta, axis=[1, 0, 0])

    def rotate_y(self, theta):
        """Rotate the shape along the y axis.

        :param theta: angle by which to rotate (in radians)
        :type theta: float

        :returns: The rotation matrix used to apply the
             transformation.
        :rtype: np.ndarray

        """
        return self.rotate(theta, axis=[0, 1, 0])

    def rotate_z(self, theta):
        """Rotate the shape along the z axis.

        :param theta: angle by which to rotate (in radians)
        :type theta: float

        :returns: The rotation matrix used to apply the
            transformation.
        :rtype: np.ndarray

        """
        return self.rotate(theta)

    @_call_on_children
    @_apply_transform
    def scale(self, sx, sy=None, sz=None):
        """Scale the shape by the given factor.

        :param sx: scale factor along the x-axis.
        :type sx: float

        :param sy: scale factor along the y-axis (defaults to None)
        :type sy: float

        :param sz: scale factor along the z-axis (defaults to None)
        :type sz: float

        :returns: The transformation matrix used to appy the
            transformation.
        :rtype: np.ndarray

        """
        if (not sy) and (not sz):
            sy = sx
            sz = sx
        elif not sz:
            sz = 1
        tmat = matrix.scale_transform(sx, sy, sz)
        return tmat

    @_call_on_children
    @_apply_transform
    def shear_x(self, theta):
        """Shear shape along the x-axis.

        :param theta: angle to shear by (in radians)
        :type theta: float

        :returns: The shear matrix used to apply the tranformation.
        :rtype: np.ndarray

        """
        shear_mat = np.identity(4)
        shear_mat[0, 1] = np.tan(theta)
        return shear_mat

    @_call_on_children
    @_apply_transform
    def shear_y(self, theta):
        """Shear shape along the y-axis.

        :param theta: angle to shear by (in radians)
        :type theta: float

        :returns: The shear matrix used to apply the transformation.
        :rtype: np.ndarray

        """
        shear_mat = np.identity(4)
        shear_mat[1, 0] = np.tan(theta)
        return shear_mat




#
# Part of p5: A Python package based on Processing
# Copyright (C) 2017-2019 Abhik Pal
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import functools
import math

import numpy as np

from ..pmath import Point
from ..pmath import curves
from ..pmath.utils import SINCOS

from .shape import PShape

from . import p5

__all__ = ['point', 'line', 'arc', 'triangle', 'quad',
           'rect', 'square', 'circle', 'ellipse', 'ellipse_mode',
           'rect_mode', 'bezier', 'curve', 'create_shape', 'draw_shape']

_rect_mode = 'CORNER'
_ellipse_mode = 'CENTER'
_shape_mode = 'CORNER'

# We use these in ellipse tessellation. The algorithm is similar to
# the one used in Processing and the we compute the number of
# subdivisions per ellipse using the following formula:
#
#    min(M, max(N, (2 * pi * size / F)))
#
# Where,
#
# - size :: is the measure of the dimensions of the circle when
#   projected in screen coordiantes.
#
# - F :: sets the minimum number of subdivisions. A smaller `F` would
#   produce more detailed circles (== POINT_ACCURACY_FACTOR)
#
# - N :: Minimum point accuracy (== MIN_POINT_ACCURACY)
#
# - M :: Maximum point accuracy (== MAX_POINT_ACCURACY)
#
MIN_POINT_ACCURACY = 20
MAX_POINT_ACCURACY = 200
POINT_ACCURACY_FACTOR = 10

def _draw_on_return(func):
    """Set shape parameters to default renderer parameters

    """
    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        s = func(*args, **kwargs)
        draw_shape(s)
        return s

    return wrapped

class Arc(PShape):
    def __init__(self, center, radii, start_angle, stop_angle,
                 attribs='open pie', fill_color='auto',
                 stroke_color='auto', stroke_weight="auto", 
                 stroke_join="auto", stroke_cap="auto", **kwargs):
        self._center = center
        self._radii = radii
        self._start_angle = start_angle
        self._stop_angle = stop_angle

        self._faces = None

        super().__init__(vertices=[], attribs=attribs, fill_color=fill_color,
                 stroke_color=stroke_color, stroke_weight=stroke_weight, 
                 stroke_join=stroke_join, stroke_cap=stroke_cap, **kwargs)
        self._tessellate()

    @property
    def _draw_vertices(self):
        if self._vertices is None:
            self._tessellate()
        return self._vertices

    @property
    def _draw_edges(self):
        if self._vertices is None:
            self._tessellate()
        n, _ = self._vertices.shape
        e = self._compute_outline_edges()
        if 'chord' in self.attribs:
            return np.concatenate([e, [[1, n - 1]]])
        return np.concatenate([e, [[0, n - 1]]])

    @property
    def _draw_faces(self):
        if self._vertices is None:
            self._tessellate()
        n, _ = self._vertices.shape
        ar = np.arange(1, n - 1).reshape((n - 2, 1))
        f = np.hstack([np.zeros((n - 2, 1)), ar, ar + 1])

        return f

    @property
    def _draw_outline_vertices(self):
        return self._compute_outline_edges()

    @property
    def _draw_outline_edges(self):
        return self._compute_outline_edges()

    def _compute_outline_edges(self):
        n, _ = self._vertices.shape
        e = np.vstack([np.arange(1, n - 1), np.arange(2, n)]).transpose()
        return e

    def _tessellate(self):
        """Generate vertex and face data using radii.
        """
        rx = self._radii[0]
        ry = self._radii[1]

        c1x = self._center[0]
        c1y = self._center[1]
        s1 = p5.renderer.transform_matrix.dot(np.array([c1x, c1y, 0, 1]))

        c2x = c1x + rx
        c2y = c1y + ry
        s2 = p5.renderer.transform_matrix.dot(np.array([c2x, c2y, 0, 1]))

        sdiff = (s2 - s1)
        size_acc = (np.sqrt(np.sum(sdiff * sdiff)) * math.pi * 2) / POINT_ACCURACY_FACTOR

        acc = min(MAX_POINT_ACCURACY, max(MIN_POINT_ACCURACY, int(size_acc)))
        inc = int(len(SINCOS) / acc)

        sclen = len(SINCOS)
        start_index = int((self._start_angle / (math.pi * 2)) * sclen)
        end_index = int((self._stop_angle / (math.pi * 2)) * sclen)

        vertices = [(c1x, c1y)]
        for idx in range(start_index, end_index, inc):
            i = idx % sclen
            vertices.append((
                c1x + rx * SINCOS[i][1],
                c1y + ry * SINCOS[i][0],
            ))
        vertices.append((
            c1x + rx * SINCOS[end_index % sclen][1],
            c1y + ry * SINCOS[end_index % sclen][0],
        ))
        self._vertices = np.array(vertices)

@_draw_on_return
def point(x, y, z=0):
    """Returns a point.

    :param x: x-coordinate of the shape.
    :type x: int or float

    :param y: y-coordinate of the shape.
    :type y: int or float

    :param z: z-coordinate of the shape (defaults to 0).
    :type z: int or float

    :returns: A point PShape.
    :rtype: PShape

    """
    return PShape([(x, y, z)], attribs='point')

@_draw_on_return
def line(p1, p2):
    """Returns a line.

    :param p1: Coordinates of the starting point of the line.
    :type p1: tuple

    :param p2: Coordinates of the end point of the line.
    :type p2: tuple

    :returns: A line PShape.
    :rtype: PShape

    """
    path = [
        Point(*p1),
        Point(*p2)
    ]
    return PShape(path, attribs='path')

@_draw_on_return
def bezier(start, control_point_1, control_point_2, stop):
    """Return a bezier path defined by two control points.

    :param start: The starting point of the bezier curve.
    :type start: tuple.

    :param control_point_1: The first control point of the bezier
        curve.
    :type control_point_1: tuple.

    :param control_point_2: The second control point of the bezier
        curve.
    :type control_point_2: tuple.

    :param stop: The end point of the bezier curve.
    :type stop: tuple.

    :returns: A bezier path.
    :rtype: PShape.

    """
    vertices = []
    steps = curves.bezier_resolution
    for i in range(steps + 1):
        t = i / steps
        p = curves.bezier_point(start, control_point_1,
                                control_point_2, stop, t)
        vertices.append(p[:3])

    return PShape(vertices, attribs='path')

@_draw_on_return
def curve(point_1, point_2, point_3, point_4):
    """Return a Catmull-Rom curve defined by four points.

    :param point_1: The first point of the curve.
    :type point_1: tuple

    :param point_2: The first point of the curve.
    :type point_2: tuple

    :param point_3: The first point of the curve.
    :type point_3: tuple

    :param point_4: The first point of the curve.
    :type point_4: tuple

    :returns: A curved path.
    :rtype: PShape

    """
    vertices = []
    steps = curves.curve_resolution
    for i in range(steps + 1):
        t = i / steps
        p = curves.curve_point(point_1, point_2, point_3, point_4, t)
        vertices.append(p[:3])

    return PShape(vertices, attribs='path')

@_draw_on_return
def triangle(p1, p2, p3):
    """Return a triangle.

    :param p1: coordinates of the first point of the triangle
    :type p1: tuple | list | p5.Vector

    :param p2: coordinates of the second point of the triangle
    :type p2: tuple | list | p5.Vector

    :param p3: coordinates of the third point of the triangle
    :type p3: tuple | list | p5.Vector

    :returns: A triangle.
    :rtype: p5.PShape
    """
    tr = PShape()
    with tr.edit():
        for pt in [p1, p2, p3]:
            tr.add_vertex(pt)
    return tr

@_draw_on_return
def quad(p1, p2, p3, p4):
    """Return a quad.

    :param p1: coordinates of the first point of the quad
    :type p1: tuple | list | p5.Vector

    :param p2: coordinates of the second point of the quad
    :type p2: tuple | list | p5.Vector

    :param p3: coordinates of the third point of the quad
    :type p3: tuple | list | p5.Vector

    :param p4: coordinates of the fourth point of the quad
    :type p4: tuple | list | p5.Vector

    :returns: A quad.
    :rtype: PShape
    """
    qd = PShape()
    with qd.edit():
        for pt in [p1, p2, p3, p4]:
            qd.add_vertex(pt)
    return qd

def rect(coordinate, *args, mode=None):
    """Return a rectangle.

    :param coordinate: Represents the lower left corner of then
        rectangle when mode is 'CORNER', the center of the rectangle
        when mode is 'CENTER' or 'RADIUS', and an arbitrary corner
        when mode is 'CORNERS'

    :type coordinate: tuple | list | p5.Vector

    :param args: For modes'CORNER' or 'CENTER' this has the form
        (width, height); for the 'RADIUS' this has the form
        (half_width, half_height); and for the 'CORNERS' mode, args
        should be the corner opposite to `coordinate`.

    :type: tuple

    :param mode: The drawing mode for the rectangle. Should be one of
        {'CORNER', 'CORNERS', 'CENTER', 'RADIUS'} (defaults to the
        mode being used by the p5.renderer.)

    :type mode: str

    :returns: A rectangle.
    :rtype: p5.PShape

    """
    if mode is None:
        mode = _rect_mode

    if mode == 'CORNER':
        corner = coordinate
        width, height = args
    elif mode == 'CENTER':
        center = Point(*coordinate)
        width, height = args
        corner = Point(center.x - width/2, center.y - height/2, center.z)
    elif mode == 'RADIUS':
        center = Point(*coordinate)
        half_width, half_height = args
        corner = Point(center.x - half_width, center.y - half_height, center.z)
        width = 2 * half_width
        height = 2 * half_height
    elif mode == 'CORNERS':
        corner = Point(*coordinate)
        corner_2, = args
        corner_2 = Point(*corner_2)
        width = corner_2.x - corner.x
        height = corner_2.y - corner.y
    else:
        raise ValueError("Unknown rect mode {}".format(mode))

    p1 = Point(*corner)
    p2 = Point(p1.x + width, p1.y, p1.z)
    p3 = Point(p2.x, p2.y + height, p2.z)
    p4 = Point(p1.x, p3.y, p3.z)
    return quad(p1, p2, p3, p4)

def square(coordinate, side_length, mode=None):
    """Return a square.

    :param coordinate: When mode is set to 'CORNER', the coordinate
        represents the lower-left corner of the square. For modes
        'CENTER' and 'RADIUS' the coordinate represents the center of
        the square.

    :type coordinate: tuple | list |p5.Vector

    :param side_length: The side_length of the square (for modes
        'CORNER' and 'CENTER') or hald of the side length (for the
        'RADIUS' mode)

    :type side_length: int or float

    :param mode: The drawing mode for the square. Should be one of
        {'CORNER', 'CORNERS', 'CENTER', 'RADIUS'} (defaults to the
        mode being used by the p5.renderer.)

    :type mode: str

    :returns: A rectangle.
    :rtype: p5.PShape

    :raises ValueError: When the mode is set to 'CORNERS'

    """
    if mode is None:
        mode = _rect_mode

    if mode == 'CORNERS':
        raise ValueError("Cannot draw square with {} mode".format(mode))
    return rect(coordinate, side_length, side_length, mode=mode)

def rect_mode(mode='CORNER'):
    """Change the rect and square drawing mode for the p5.renderer.

    :param mode: The new mode for drawing rects. Should be one of
        {'CORNER', 'CORNERS', 'CENTER', 'RADIUS'}. This defaults to
        'CORNER' so calling rect_mode without parameters will reset
        the sketch's rect mode.
    :type mode: str

    """
    global _rect_mode
    _rect_mode = mode

@_draw_on_return
def arc(coordinate, width, height, start_angle, stop_angle,
        mode='OPEN PIE', ellipse_mode=None):
    """Return a ellipse.

    :param coordinate: Represents the center of the arc when mode
        is 'CENTER' (the default) or 'RADIUS', the lower-left corner
        of the ellipse when mode is 'CORNER'.

    :rtype coordinate: 3-tuple

    :param width: For ellipse modes 'CORNER' or 'CENTER' this
        represents the width of the the ellipse of which the arc is a
        part. Represents the x-radius of the parent ellipse when
        ellipse mode is 'RADIUS

    :type width: float

    :param height: For ellipse modes 'CORNER' or 'CENTER' this
        represents the height of the the ellipse of which the arc is a
        part. Represents the y-radius of the parent ellipse when
        ellipse mode is 'RADIUS

    :type height: float

    :param mode: The mode used to draw an arc can be some combination
        of {'OPEN', 'CHORD', 'PIE'} separated by spaces. For instance,
        'OPEN PIE', etc (defaults to 'OPEN PIE')

    :type mode: str

    :param ellipse_mode: The drawing mode used for the ellipse. Should be one of
        {'CORNER', 'CENTER', 'RADIUS'} (defaults to the
        mode being used by the p5.renderer.)

    :type mode: str

    :returns: An arc.
    :rtype: Arc

    """
    amode = mode

    if ellipse_mode is None:
        emode = _ellipse_mode
    else:
        emode = ellipse_mode

    if emode == 'CORNER':
        corner = Point(*coordinate)
        dim = Point(width/2, height/2)
        center = (corner.x + dim.x, corner.y + dim.y, corner.z)
    elif emode == 'CENTER':
        center = Point(*coordinate)
        dim = Point(width / 2, height / 2)
    elif emode == 'RADIUS':
        center = Point(*coordinate)
        dim = Point(width, height)
    else:
        raise ValueError("Unknown arc mode {}".format(emode))
    return Arc(center, dim, start_angle, stop_angle, attribs=amode)

def ellipse(coordinate, *args, mode=None):
    """Return a ellipse.

    :param coordinate: Represents the center of the ellipse when mode
        is 'CENTER' (the default) or 'RADIUS', the lower-left corner
        of the ellipse when mode is 'CORNER' or, and an arbitrary
        corner when mode is 'CORNERS'.

    :type coordinate: 3-tuple

    :param args: For modes'CORNER' or 'CENTER' this has the form
        (width, height); for the 'RADIUS' this has the form
        (x_radius, y_radius); and for the 'CORNERS' mode, args
        should be the corner opposite to `coordinate`.

    :type: tuple

    :param mode: The drawing mode for the ellipse. Should be one of
        {'CORNER', 'CORNERS', 'CENTER', 'RADIUS'} (defaults to the
        mode being used by the p5.renderer.)

    :type mode: str

    :returns: An ellipse
    :rtype: Arc

    """
    if mode is None:
        mode = _ellipse_mode

    if mode == 'CORNERS':
        corner = Point(*coordinate)
        corner_2, = args
        corner_2 = Point(*corner_2)
        width = corner_2.x - corner.x
        height = corner_2.y - corner.y
        mode = 'CORNER'
    else:
        width, height = args
    return arc(coordinate, width, height, 0, math.pi * 2, 'CHORD', mode)

def circle(coordinate, radius, mode=None):
    """Return a circle.

    :param coordinate: Represents the center of the ellipse when mode
        is 'CENTER' (the default) or 'RADIUS', the lower-left corner
        of the ellipse when mode is 'CORNER' or, and an arbitrary
        corner when mode is 'CORNERS'.

    :type coordinate: 3-tuple

    :param radius: For modes'CORNER' or 'CENTER' this actually
        represents the diameter; for the 'RADIUS' this represents the
        radius.

    :type: tuple

    :param mode: The drawing mode for the ellipse. Should be one of
        {'CORNER', 'CORNERS', 'CENTER', 'RADIUS'} (defaults to the
        mode being used by the p5.renderer.)

    :type mode: str

    :returns: A circle.
    :rtype: Ellipse

    :raises ValueError: When mode is set to 'CORNERS'

    """
    if mode is None:
        mode = _ellipse_mode

    if mode == 'CORNERS':
        raise ValueError("Cannot create circle in CORNERS mode")
    return ellipse(coordinate, radius, radius, mode=mode)

def ellipse_mode(mode='CENTER'):
    """Change the ellipse and circle drawing mode for the p5.renderer.

    :param mode: The new mode for drawing ellipses. Should be one of
        {'CORNER', 'CORNERS', 'CENTER', 'RADIUS'}. This defaults to
        'CENTER' so calling ellipse_mode without parameters will reset
        the sketch's ellipse mode.
    :type mode: str

    """
    global _ellipse_mode
    _ellipse_mode = mode

def draw_shape(shape, pos=(0, 0, 0)):
    """Draw the given shape at the specified location.

    :param shape: The shape that needs to be drawn.
    :type shape: p5.PShape

    :param pos: Position of the shape
    :type pos: tuple | Vector

    """
    p5.renderer.render(shape)
    for child_shape in shape.children:
        draw_shape(child_shape)

def create_shape(kind=None, *args, **kwargs):
    """Create a new PShape

    Note :: A shape created using this function is *not* visible by
        default. Please make the shape visible by setting the shapes's
        `visible` attribute to true.

    :param kind: Type of shape. When left unspecified a generic PShape
        is returned (which can be edited later). Valid values for
        `kind` are: { 'point', 'line', 'triangle', 'quad', 'rect',
        'ellipse', 'arc', }

    :type kind: None | str

    :param args: Initial arguments to be passed to the shape creation
        function (only applied when `kind` is *not* None)

    :param kwargs: Initial keyword arguments to be passed to the shape
        creation function (only applied when `kind` is *not* None)

    :returns: The requested shape
    :rtype: p5.PShape

    """

    # TODO: add 'box', 'sphere' support
    valid_values = { None, 'point', 'line', 'triangle', 'quad',
                     'rect', 'square', 'ellipse', 'circle', 'arc', }

    def empty_shape(*args, **kwargs):
        return PShape()

    shape_map = {
        'arc': arc,
        'circle': circle,
        'ellipse': ellipse,
        'line': line,
        'point': point,
        'quad': quad,
        'rect': rect,
        'square': square,
        'triangle': triangle,
        None: empty_shape,
    }

    # kwargs['visible'] = False
    return shape_map[kind](*args, **kwargs)



#
# Part of p5: A Python package based on Processing
# Copyright (C) 2017-2019 Abhik Pal
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

"""Base module for a sketch."""

import builtins

from PIL import Image
from vispy import app

from ..core import p5

from .events import KeyEvent
from .events import MouseEvent
from .events import handler_names


def _dummy(*args, **kwargs):
    """Eat all arguments, do nothing.
    """
    pass

class Sketch(app.Canvas):
    """The main sketch instance.

    :param setup_method: Setup method for the sketch. This is run
        exactly once for each run of the sketch.
    :type setup_method: function

    :param draw_method: Draw method for the sketch which keeps running
        indefinitely.
    :type draw_method: function

    :param handlers: Dictionary containing the event handlers for the
        sketch. By default, maps to an empty dict.
        nothing.
    :type handlers: { str: function }

    :param frame_rate:
    :type frame_rate: int

    """
    def __init__(self, setup_method, draw_method,
                 handlers=dict(), frame_rate=60):
        app.Canvas.__init__(
            self,
            title=builtins.title,
            size=(builtins.width, builtins.height),
            keys='interactive',
            resizable=False,
        )

        self.setup_method = setup_method
        self.draw_method = draw_method

        self.looping = None
        self.redraw = None
        self.setup_done = False
        self.timer = app.Timer(1.0 / frame_rate, connect=self.on_timer)

        self.handlers = dict()
        for handler_name in handler_names:
            self.handlers[handler_name] = handlers.get(handler_name, _dummy)

        self.handler_queue = []

        self._save_fname = 'screen'
        self._save_fname_num = 0
        self._save_flag = False

        p5.renderer.initialize_renderer()
        p5.renderer.clear()

    def on_timer(self, event):
        self.measure_fps(callback=lambda _: None)
        builtins.frame_rate = round(self.fps, 2)

        with p5.renderer.draw_loop():
            builtins.frame_count += 1
            if not self.setup_done:
                self.setup_method()
                self.setup_done = True
                self.show(visible=True)
                if self.redraw is None:
                    self.redraw = False
                if self.looping is None:
                    self.looping = True
            elif self.looping:
                self.draw_method()
            elif self.redraw:
                self.draw_method()
                self.redraw = False

            while len(self.handler_queue) != 0:
                function, event = self.handler_queue.pop(0)
                event._update_builtins()
                function(event)

        if self._save_flag:
            self._save_buffer()
        self.update()

    def _save_buffer(self):
        """Save the renderer buffer to the given file.
        """
        img_data = p5.renderer.fbuffer.read(mode='color', alpha=False)
        img = Image.fromarray(img_data)
        img.save(self._save_fname)
        self._save_flag = False

    def screenshot(self, filename):
        self.queue_screenshot(filename)
        p5.renderer.flush_geometry()
        self._save_buffer()

    def queue_screenshot(self, filename):
        """Save the current frame
        """
        fname_split = filename.split('.')
        ext = '.' + fname_split[-1]
        stem = '.'.join(fname_split[:-1])
        self._save_fname = stem + str(self._save_fname_num).zfill(4) + ext
        self._save_fname_num = self._save_fname_num + 1
        self._save_flag = True

    def on_close(self, event):
        exit()

    def on_draw(self, event):
        pass

    def on_resize(self, event):
        p5.renderer.reset_view()
        with p5.renderer.draw_loop():
            p5.renderer.clear()

    def _enqueue_event(self, handler_name, event):
        event._update_builtins()
        self.handler_queue.append((self.handlers[handler_name], event))

    def on_key_press(self, event):
        kev = KeyEvent(event, active=True)
        self._enqueue_event('key_pressed', kev)

    def on_key_release(self, event):
        kev = KeyEvent(event)
        self._enqueue_event('key_released', kev)
        if not (event.text is ''):
            self._enqueue_event('key_typed', kev)

    def on_mouse_press(self, event):
        mev = MouseEvent(event, active=True)
        self._enqueue_event('mouse_pressed', mev)

    def on_mouse_double_click(self, event):
        mev = MouseEvent(event)
        self._enqueue_event('mouse_double_clicked', mev)

    def on_mouse_release(self, event):
        mev = MouseEvent(event)
        self._enqueue_event('mouse_released', mev)
        self._enqueue_event('mouse_clicked', mev)

    def on_mouse_move(self, event):
        mev = MouseEvent(event, active=builtins.mouse_is_pressed)
        self._enqueue_event('mouse_moved', mev)
        if builtins.mouse_is_pressed:
            self._enqueue_event('mouse_dragged', mev)

    def on_mouse_wheel(self, event):
        mev = MouseEvent(event, active=builtins.mouse_is_pressed)
        self._enqueue_event('mouse_wheel', mev)

    # def on_touch(self, event):
    #     self._enqueue_event('touch', event)

    # def on_stylus(self, event):
    #     self._enqueue_event('stylus', event)



#
# Part of p5: A Python package based on Processing
# Copyright (C) 2017-2019 Abhik Pal
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

from contextlib import contextmanager

import numpy as np

from ..pmath import matrix

from . import p5

__all__ = ['push_matrix', 'pop_matrix', 'reset_transforms', 
           'translate', 'rotate', 'rotate_x', 'rotate_y', 
           'rotate_z', 'scale', 'shear_x', 'shear_y', 
           'camera', 'frustum', 'ortho', 'perspective',
           'print_matrix', 'reset_matrix', 'apply_matrix']

@contextmanager
def push_matrix():
    previous_matrix = p5.renderer.transform_matrix.copy()
    try:
        yield previous_matrix
    finally:
        p5.renderer.transform_matrix = previous_matrix

def pop_matrix():
    push_matrix()

def reset_transforms():
    """Reset all transformations to their default state.

    """
    p5.renderer.transform_matrix = np.identity(4)

def translate(x, y, z=0):
    """Translate the display origin to the given location.

    :param x: The displacement amount in the x-direction (controls the
        left/right displacement)

    :type x: int

    :param y: The displacement amount in the y-direction (controls the
        up/down displacement)

    :type y: int

    :param z: The displacement amount in the z-direction (0 by
        default). This controls the displacement away-from/towards the
        screen.
    :type z: int

    :returns: The translation matrix applied to the transform matrix.
    :rtype: np.ndarray

    """
    tmat = matrix.translation_matrix(x, y, z)
    p5.renderer.transform_matrix = p5.renderer.transform_matrix.dot(tmat)
    return tmat

def rotate(theta, axis=np.array([0, 0, 1])):
    """Rotate the display by the given angle along the given axis.

    :param theta: The angle by which to rotate (in radians)
    :type theta: float

    :param axis: The axis along which to rotate (defaults to the z-axis)
    :type axis: np.ndarray or list

    :returns: The rotation matrix used to apply the transformation.
    :rtype: np.ndarray

   """
    axis = np.array(axis[:])
    tmat = matrix.rotation_matrix(axis, theta)
    p5.renderer.transform_matrix = p5.renderer.transform_matrix.dot(tmat)
    return tmat

def rotate_x(theta):
    """Rotate the view along the x axis.

    :param theta: angle by which to rotate (in radians)
    :type theta: float

    :returns: The rotation matrix used to apply the transformation.
    :rtype: np.ndarray

    """
    rotate(theta, axis=np.array([1, 0, 0]))

def rotate_y(theta):
    """Rotate the view along the y axis.

    :param theta: angle by which to rotate (in radians)
    :type theta: float

    :returns: The rotation matrix used to apply the transformation.
    :rtype: np.ndarray

   """
    rotate(theta, axis=np.array([0, 1, 0]))

def rotate_z(theta):
    """Rotate the view along the z axis.

    :param theta: angle by which to rotate (in radians)
    :type theta: float

    :returns: The rotation matrix used to apply the transformation.
    :rtype: np.ndarray

   """
    rotate(theta, axis=np.array([0, 0, 1]))

def scale(sx, sy=None, sz=None):
    """Scale the display by the given factor.

    :param sx: scale factor along the x-axis.
    :type sx: float

    :param sy: scale factor along the y-axis (defaults to None)
    :type sy: float

    :param sz: scale factor along the z-axis (defaults to None)
    :type sz: float

    :returns: The transformation matrix used to appy the transformation.
    :rtype: np.ndarray
    """
    if (not sy) and (not sz):
        sy = sx
        sz = sx
    elif not sz:
        sz = 1
    tmat = matrix.scale_transform(sx, sy, sz)
    p5.renderer.transform_matrix = p5.renderer.transform_matrix.dot(tmat)
    return tmat

def apply_matrix(transform_matrix):
    """Apply the given matrix to the sketch's transform matrix..

    :param transform_matrix: The new transform matrix.
    :type transform_matrix: np.ndarray (or a 4×4 list)
    """
    tmatrix = np.array(transform_matrix)
    p5.renderer.transform_matrix = p5.renderer.transform_matrix.dot(tmatrix)

def reset_matrix():
    """Reset the current transform matrix.
    """
    p5.renderer.transform_matrix = np.identity(4)

def print_matrix():
    """Print the transform matrix being used by the sketch.
    """
    print(p5.renderer.transform_matrix)

def shear_x(theta):
    """Shear display along the x-axis.

    :param theta: angle to shear by (in radians)
    :type theta: float

    :returns: The shear matrix used to apply the tranformation.
    :rtype: np.ndarray

    """
    shear_mat = np.identity(4)
    shear_mat[0, 1] = np.tan(theta)
    p5.renderer.transform_matrix = p5.renderer.transform_matrix.dot(shear_mat)
    return shear_mat

def shear_y(theta):
    """Shear display along the y-axis.

    :param theta: angle to shear by (in radians)
    :type theta: float

    :returns: The shear matrix used to apply the transformation.
    :rtype: np.ndarray

    """
    shear_mat = np.identity(4)
    shear_mat[1, 0] = np.tan(theta)
    p5.renderer.transform_matrix = p5.renderer.transform_matrix.dot(shear_mat)
    return shear_mat

def camera(position, target_position, up_vector):
    """Sets the camera position for a 3D sketch. 
    Parameters for this function define the position for 
    the camera, the center of the sketch (where the 
    camera is pointing), and an up direction (the 
    orientation of the camera).

    When called with no arguments, this function 
    creates a default camera equivalent to 
    camera(
        (0, 0, (height/2.0) / tan(PI*30.0 / 180.0)),
        (0, 0, 0),
        (0, 1, 0)
        )

    :param position: camera position coordinates
    :type position: tuple

    :param target_position: target position of camera in world coordinates 
    :type target_position: tuple

    :param up_vector: up direction vector for the camera
    :type up_vector: tuple

    """
    p5.renderer.lookat_matrix = matrix.look_at(
        np.array(position), 
        np.array(target_position), 
        np.array(up_vector))

def perspective(fovy, aspect, near, far):
    """
    Sets a perspective projection for the camera in a 3D sketch.  

    :param fovy: camera frustum vertical field of view, from bottom to top of view, in angleMode units
    :type fovy: float

    :param aspect: camera frustum aspect ratio
    :type aspect: float

    :param near: frustum near plane length
    :type near: float

    :param far: frustum far plane length
    :type far: float   
    """
    p5.renderer.projection_matrix = matrix.perspective_matrix(
            fovy,
            aspect,
            near,
            far
        )

def ortho(left, right, bottom, top, near, far):
    """
    Sets an orthographic projection for the camera in
    a 3D sketch and defines a box-shaped viewing frustum
    within which objects are seen.

    :param left: camera frustum left plane
    :type left: float

    :param right: camera frustum right plane
    :type right: float

    :param bottom: camera frustum bottom plane
    :type bottom: float

    :param top: camera frustum top plane
    :type top: float

    :param near: camera frustum near plane
    :type near: float

    :param far: camera frustum far plane
    :type far: float   
    """
    p5.renderer.projection_matrix = np.array([
        [2/(right - left), 0, 0, -(right + left)/(right - left)],
        [0, 2/(top - bottom), 0, -(top + bottom)/(top - bottom)],
        [0, 0, -2/(far - near), -(far + near)/(far - near)],
        [0, 0, 0, 1],
        ])

def frustum():
    raise NotImplementedError



import unittest
import numpy as np

from p5.core.shape import PShape
from p5.core.color import Color
from p5.pmath import PI

vertices = [(0, 0), (1, 0), (1, 1), (0, 1)]

quad = PShape(vertices=vertices, fill_color=Color(255),
        stroke_color=Color(0), stroke_weight=2,
        stroke_join=1, stroke_cap=1, attribs="closed")

contour_shape = PShape(vertices=[(0, 0), (100, 0), (100, 100), (0, 100)], 
        fill_color=Color(255), stroke_color=Color(0), stroke_weight=2,
        stroke_join=1, stroke_cap=1, attribs="closed",
        contour=[[25, 25], [75, 25], [75, 75], [25, 75]])

class TestPShape(unittest.TestCase):
    def test_properties(self):
        self.assertTrue(np.array_equal(
            quad.vertices, 
            np.array([(0, 0), (1, 0), (1, 1), (0, 1)])))
        self.assertEqual(quad._fill, Color(255))
        self.assertEqual(quad._stroke, Color(0))
        self.assertEqual(quad._stroke_weight, 2)
        self.assertEqual(quad._stroke_cap, 1)
        self.assertEqual(quad._stroke_join, 1)

    def test_transforms(self):
        quad.translate(100, 100, 100)
        self.assertTrue(np.array_equal(
            quad._matrix, 
            np.array([
                [1.0, 0.0, 0.0, 100.0],
                [0.0, 1.0, 0.0, 100.0],
                [0.0, 0.0, 1.0, 100.0],
                [0.0, 0.0, 0.0, 1.0],
                ])))
        quad.reset_matrix()


        quad.rotate_x(PI/2)
        self.assertTrue(np.allclose(
            quad._matrix, 
            np.array([
                [1.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, -1.0, 0.0],
                [0.0, 1.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 1.0],
                ])))
        quad.reset_matrix()


        quad.rotate_y(PI/2)
        self.assertTrue(np.allclose(
            quad._matrix, 
            np.array([
                [0.0, 0.0, 1.0, 0.0],
                [0.0, 1.0, 0.0, 0.0],
                [-1.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 1.0],
                ])))
        quad.reset_matrix()


        quad.rotate_z(PI/2)
        self.assertTrue(np.allclose(
            quad._matrix, 
            np.array([
                [0.0, -1.0, 0.0, 0.0],
                [1.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 1.0, 0.0],
                [0.0, 0.0, 0.0, 1.0],
                ])))
        quad.reset_matrix()


        quad.scale(2, 2)
        self.assertTrue(np.allclose(
            quad._matrix, 
            np.array([
                [2.0, 0.0, 0.0, 0.0],
                [0.0, 2.0, 0.0, 0.0],
                [0.0, 0.0, 1.0, 0.0],
                [0.0, 0.0, 0.0, 1.0],
                ])))
        quad.reset_matrix()


        quad.shear_x(PI/4)
        self.assertTrue(np.allclose(
            quad._matrix, 
            np.array([
                [1.0, 1.0, 0.0, 0.0],
                [0.0, 1.0, 0.0, 0.0],
                [0.0, 0.0, 1.0, 0.0],
                [0.0, 0.0, 0.0, 1.0],
                ])))
        quad.reset_matrix()


        quad.shear_y(PI/4)
        self.assertTrue(np.allclose(
            quad._matrix, 
            np.array([
                [1.0, 0.0, 0.0, 0.0],
                [1.0, 1.0, 0.0, 0.0],
                [0.0, 0.0, 1.0, 0.0],
                [0.0, 0.0, 0.0, 1.0],
                ])))
        quad.reset_matrix()

    def test_edges(self):
        self.assertTrue(
            np.array_equal(quad._compute_poly_edges(), 
            np.array([[0, 1], [1, 2], [2, 3], [3, 0]])))
        self.assertTrue(
            np.array_equal(quad._compute_outline_edges(), 
            np.array([[0, 1], [1, 2], [2, 3]])))

    def test_triangulation(self):
        quad._retriangulate()
        self.assertTrue(
            np.array_equal(quad._tri_edges, 
            np.array([[1, 0], [2, 1], [3, 2], [0, 3]])))

        self.assertTrue(
            np.array_equal(quad._tri_faces, 
            np.array([[3, 0, 1], [1, 2, 3]])))

        self.assertTrue(
            np.array_equal(quad._tri_vertices, 
            np.array([[0.0, 0.0], [1.0, 0.0], [1.0, 1.0], [0.0, 1.0]])))

        contour_shape._retriangulate()
        self.assertTrue(
            np.array_equal(contour_shape._tri_edges, 
            np.array([[1, 0], [2, 1], [3, 2], [0, 3], [4, 5], [6, 5], [7, 6], [4, 7]])))

        self.assertTrue(
            np.array_equal(contour_shape._tri_faces, 
            np.array([[0, 4, 7], [4, 0, 1], [3, 7, 6], [7, 3, 0], [5, 1, 2], [1, 5, 4], [6, 2, 3], [2, 6, 5]])))

        self.assertTrue(
            np.array_equal(contour_shape._tri_vertices, 
            np.array([[0.0, 0.0], [100.0, 0.0], [100.0, 100.0], [0.0, 100.0], [25.0, 25.0], [75.0, 25.0], [75.0, 75.0], [25.0, 75.0]])))  


if __name__ == "__main__":
    unittest.main()


#
# Part of p5: A Python package based on Processing
# Copyright (C) 2017-2019 Abhik Pal
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import random

from ..pmath import constrain
from .utils import SINCOS_LENGTH
from .utils import PRE_COS

__all__ = [
    # PERLIN NOISE FUNCTIONS
    'noise', 'noise_detail', 'noise_seed',

    # RANDOM NUMBER GENERATION
    'random_uniform', 'random_gaussian', 'random_seed'
]

# Most of the perlin noise code is based on the original Processing
# implementation of the noise function. toxi (+ other folks) put a
# bunch of comments on the Processing version, and I've added them
# here too for context.
#
# All commets that were ported from Processing are prefixed with
# "#P:"
#
#         --- abhikpal (2017-08-04)

#P: PERLIN NOISE
#P:
#P: [toxi 040903]
#P: octaves and amplitude amount per octave are now user controlled via
#P: the noiseDetail() function.
#P:
#P: [toxi 030902]
#P: cleaned up code and now using bagel's cosine table to speed up
#P:
#P: [toxi 030901]
#P: implementation by the german demo group farbrausch as used in their
#P: demo "art": http://www.farb-rausch.de/fr010src.zip

#P: Default to medium smooth
PERLIN_OCTAVES = 4

#P: 50% redution per octave
PERLIN_FALLOFF = 0.5

PERLIN_YWRAPB = 4
PERLIN_YWRAP = 1 << PERLIN_YWRAPB
PERLIN_ZWRAPB = 8
PERLIN_ZWRAP = 1 << PERLIN_ZWRAPB
PERLIN_SIZE = 4095

#P: [toxi 031112]
#P: new vars needed due to recent change of cos table in PGraphics
PERLIN_COS_TABLE = PRE_COS
PERLIN_TWO_PI = SINCOS_LENGTH
PERLIN_PI = PERLIN_TWO_PI
PERLIN_PI >>= 1

PERLIN = None

def noise(x, y=0, z=0):
    """Return perlin noise value at the given location.

    :param x: x-coordinate in noise space.
    :type x: float

    :param y: y-coordinate in noise space.
    :type y: float

    :param z: z-coordinate in noise space.
    :type z: float

    :returns: The perlin noise value.
    :rtype: float

    """
    # TODO (abhikpal, 2017-08-04)
    #
    # REFACTOR THIS MESS.

    global PERLIN

    #P: [toxi 031112]
    #P: now adjusts to the size of the cosLUT used via
    #P: the new variables, defined above
    def noise_fsc(i):
        #P: using bagel's cosine table instead
        return 0.5 * (1 - PERLIN_COS_TABLE[int(i * PERLIN_PI) % PERLIN_TWO_PI])

    #P: [toxi 031112]
    #P: noise broke due to recent change of cos table in PGraphics
    #P: this will take care of it
    if PERLIN is None:
        PERLIN = [random.random() for _ in range(PERLIN_SIZE + 1)]

    x = (-1 * x) if x < 0 else x
    xi = int(x)
    xf = x - xi

    y = (-1 * y) if y < 0 else y
    yi = int(y)
    yf = y - yi

    z = (-1 * z) if z < 0 else z
    zi = int(z)
    zf = z - zi

    r = 0
    ampl = 0.5

    for i in range(PERLIN_OCTAVES):
        rxf = noise_fsc(xf)
        ryf = noise_fsc(yf)

        of = int(xi + (yi << PERLIN_YWRAPB) + (zi << PERLIN_ZWRAPB))
        n1 = PERLIN[of % PERLIN_SIZE]
        n1 += rxf * (PERLIN[(of + 1) % PERLIN_SIZE] - n1)
        n2 = PERLIN[(of + PERLIN_YWRAP) % PERLIN_SIZE]
        n2 += rxf * (PERLIN[(of + PERLIN_YWRAP + 1) & PERLIN_SIZE] - n2)
        n1 += ryf * (n2 - n1)

        of += PERLIN_ZWRAP
        n2 = PERLIN[of & PERLIN_SIZE]
        n2 += rxf * (PERLIN[(of + 1) % PERLIN_SIZE] - n2)
        n3 = PERLIN[(of + PERLIN_YWRAP) % PERLIN_SIZE]
        n3 += rxf * (PERLIN[(of + PERLIN_YWRAP + 1) % PERLIN_SIZE] - n3)

        n2 += ryf * (n3 - n2)
        n1 += noise_fsc(zf) * (n2 - n1)

        r += n1 * ampl
        ampl *= PERLIN_FALLOFF

        xi *= 2
        xf *= 2

        yi *= 2
        yf *= 2

        zi *= 2
        zf *= 2

        if xf >= 1:
            xi = xi + 1
            xf = xf - 1

        if yf >= 1:
            yi = yi + 1
            yf = yf - 1

        if zf >= 1:
            zi = zi + 1
            zf = zf - 1

    return r

def noise_detail(octaves=4, falloff=0.5):
    """Adjust the level of noise detail produced by noise().

    :param octaves: The number of octaves to compute the noise for
        (defaults to 4).
    :type octaves: int

    :param falloff:
    :type falloff: float

    :note: For :code:`falloff` values greater than 0.5,
        :code:`noise()` will return values greater than 1.0.

    """
    global PERLIN_OCTAVES
    global PERLIN_FALLOFF

    if octaves > 0:
        PERLIN_OCTAVES = octaves
    PERLIN_FALLOFF = constrain(falloff, 0, 1)

def noise_seed(seed):
    """Set the seed value for :code:`noise()`

    By default :code:`noise()` produes different values each time the
    sketch is run. Setting the :code:`seed` parameter to a constant
    will make :code:`noise()` return the same values each time the
    sketch is run.

    :param seed: The required seed value.
    :type seed: int

    """
    global PERLIN
    random_seed(seed)
    PERLIN = None

def random_uniform(high=1, low=0):
    """Return a uniformly sampled random number.

    :param high: The upper limit on the random value (defaults to 1).
    :type high: float

    :param low: The lowe limit on the random value (defaults to 0).
    :type low: float

    :returns: A random number between :code:`low` and :code:`high`.
    :rtype: float

    """
    return random.uniform(low, high)

def random_gaussian(mean=0, std_dev=1):
    """Return a normally sampled random number.

    :param mean: The mean value to be used for the normal distribution
        (defaults to 0).
    :type mean: float

    :param std_dev: The standard deviation to be used for the normal
        distribution (defaults to 1).
    :type std_dev: float

    :returns: A random number selected from a normal distribution with
        the given :code:`mean` and :code:`std_dev`.
    :rtype: float

    """
    return random.gauss(mean, std_dev)

def random_seed(seed):
    """Set the seed used to generate random numbers.

    :param seed: The required seed value.
    :type seed: int
    """
    random.seed(seed)
